from __future__ import annotations

import os
import re
import subprocess
import tempfile
import time
import warnings
from collections import defaultdict
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Optional

import numpy as np
from addict import Dict
from cadet import Cadet as CadetAPI

from CADETProcess import CADETProcessError, SimulationResults, settings
from CADETProcess.dataStructure import (
    Bool,
    ParameterWrapper,
    Structure,
    Switch,
    UnsignedFloat,
    UnsignedInteger,
)
from CADETProcess.processModel import (
    BindingBaseClass,
    Cstr,
    FlowSheet,
    Inlet,
    LumpedRateModelWithoutPores,
    NoBinding,
    NoDiscretization,
    NoReaction,
    Process,
    ReactionBaseClass,
    TubularReactor,
    UnitBaseClass,
)
from CADETProcess.solution import (
    SolutionBulk,
    SolutionIO,
    SolutionParticle,
    SolutionSolid,
    SolutionVolume,
)

from .simulator import SimulatorBase

__all__ = [
    "Cadet",
    "ModelSolverParameters",
    "UnitParameters",
    "AdsorptionParameters",
    "ReactionParameters",
    "SolverParameters",
    "SolverTimeIntegratorParameters",
    "ReturnParameters",
    "SensitivityParameters",
]


class Cadet(SimulatorBase):
    """
    CADET class for running a simulation for given process objects.

    Attributes
    ----------
    install_path: str
        Path to the root of the CADET installation
    time_out : UnsignedFloat
        Maximum duration for simulations
    model_solver_parameters : ModelSolverParameters
        Container for solver parameters
    unit_discretization_parameters : UnitDiscretizationParameters
        Container for unit discretization parameters
    discretization_weno_parameters : DiscretizationWenoParameters
        Container for weno discretization parameters in units
    adsorption_consistency_solver_parameters : ConsistencySolverParameters
        Container for consistency solver parameters
    solver_parameters : SolverParameters
        Container for general solver settings
    time_integrator_parameters : SolverTimeIntegratorParameters
        Container for time integrator parameters
    return_parameters : ReturnParameters
        Container for return information of the system

    ..todo::
        Implement method for loading CADET file that have not been generated
        with CADETProcess and create Process

    See Also
    --------
    ReturnParameters
    ModelSolverParameters
    SolverParameters
    SolverTimeIntegratorParameters
    CadetAPI
    """

    timeout = UnsignedFloat()
    use_dll = Bool(default=False)
    _force_constant_flow_rate = False

    def __init__(
        self,
        install_path: Optional[str] = None,
        temp_dir: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize Cadet implementation of Simulator Base."""
        super().__init__(*args, **kwargs)

        if install_path is None:
            self.install_path = CadetAPI.autodetect_cadet()
        else:
            self.install_path = install_path

        self.model_solver_parameters = ModelSolverParameters()
        self.solver_parameters = SolverParameters()
        self.time_integrator_parameters = SolverTimeIntegratorParameters()

        self.return_parameters = ReturnParameters()
        self.sensitivity_parameters = SensitivityParameters()

        if temp_dir is None:
            temp_dir = settings.temp_dir / "simulation_files"
        self.temp_dir = temp_dir

    @property
    def temp_dir(self) -> dict:
        """Return temp dir."""
        if not self._temp_dir.exists():
            self._temp_dir.mkdir(exist_ok=True, parents=True)
        return self._temp_dir

    @temp_dir.setter
    def temp_dir(self, temp_dir: str) -> None:
        self._temp_dir = temp_dir

    def locks_process(func: Callable) -> Callable:
        """Lock process to enable caching."""

        @wraps(func)
        def wrapper(self: Any, process: Process, *args: Any, **kwargs: Any) -> Any:
            locked_process = False

            if not process.lock:
                process.lock = True
                locked_process = True

            results = func(self, process, *args, **kwargs)

            if locked_process:
                process.lock = False

            return results

        return wrapper

    def check_cadet(self) -> bool:
        """
        Check if CADET installation can run a basic LWE example.

        Returns
        -------
        bool
            True if the test simulation completed successfully, False otherwise.

        Raises
        ------
        CADETProcessError
            If the simulation fails, an exception is raised with the error message.

        Notes
        -----
        This method tests the CADET installation by executing a basic LWE (Load, Wash,
        Elute) example. It creates a CADET model using the LWE data stored in an HDF5
        file and runs the simulation. After the simulation, it checks the return code to
        determine if the test was successful.
        """
        lwe_hdf5_path = Path(self.temp_dir) / "LWE.h5"

        cadet = self.get_new_cadet_instance()

        cadet.create_lwe(lwe_hdf5_path)

        # Check for CADET-Python >= v1.1, which introduced the .run_simulation interface.
        # If it's not present, assume CADET-Python <= 1.0.4 and use the old .run_load()
        # interface. This check can be removed at some point in the future.
        if hasattr(cadet, "run_simulation"):
            return_information = cadet.run_simulation(timeout=self.timeout)
        else:
            return_information = cadet.run_load(timeout=self.timeout)

        cadet.delete_file()

        if return_information.return_code != 0:
            raise CADETProcessError(return_information.error_message)

        print("Test simulation completed successfully")

        return True

    def get_tempfile_name(self) -> str:
        """Return name of tempfile."""
        f = next(tempfile._get_candidate_names())
        return self.temp_dir / f"{f}.h5"

    @locks_process
    def _run(
        self,
        process: Process,
        cadet: CadetAPI = None,
        file_path: Optional[os.PathLike] = None,
    ) -> SimulationResults:
        """
        Interface to the solver run function.

        The configuration is extracted from the process object and then saved
        as a temporary .h5 file. After termination, the same file is processed
        and the results are returned.

        Cadet Return information:
        - 0: pass (everything allright)
        - 1: Standard Error
        - 2: IO Error
        - 3: Solver Error

        Parameters
        ----------
        process : Process
            process to be simulated
        cadet : Cadet, Optional, Default=None
            Cadet instance. If None, new Cadet instance will be initialized.
        file_path : os.PathLike
            File path where to save Cadet files.

        Returns
        -------
        results : SimulationResults
            Simulation results including process and solver configuration.

        Raises
        ------
        TypeError
            If process is not instance of Process

        See Also
        --------
        get_process_config
        get_simulation_results
        """
        if not isinstance(process, Process):
            raise TypeError("Expected Process")

        if cadet is None:
            cadet = self.get_new_cadet_instance()

        cadet.root = self.get_process_config(process)

        if not self.use_dll:
            if file_path is None:
                cadet.filename = self.get_tempfile_name()
            else:
                cadet.filename = file_path
            cadet.save()

        try:
            start = time.time()

            # Check for CADET-Python >= v1.1, which introduced the .run_simulation interface.
            # If it's not present, assume CADET-Python <= 1.0.4 and use the old .run_load()
            # interface. This check can be removed at some point in the future.
            if hasattr(cadet, "run_simulation"):
                return_information = cadet.run_simulation(timeout=self.timeout)
            else:
                return_information = cadet.run_load()
            elapsed = time.time() - start
        except subprocess.TimeoutExpired:
            raise CADETProcessError("Simulator timed out") from None
        finally:
            if not self.use_dll and file_path is None:
                os.remove(cadet.filename)

        if return_information.return_code != 0:
            self.logger.error(
                f"Simulation of {process.name} with parameters {process.config} failed."
            )
            raise CADETProcessError(
                f"CADET Error: Simulation failed with {return_information.error_message}"
            ) from None

        try:
            results = self.get_simulation_results(
                process, cadet, elapsed, return_information
            )
        except TypeError:
            raise CADETProcessError("Unexpected error reading SimulationResults.")

        return results

    def get_cadet_version(self) -> tuple[str, str]:
        """
        Get version and branch name of the currently instanced CADET build.

        Returns
        -------
        tuple[str, str]
            The CADET version as x.x.x and the branch name.

        Raises
        ------
        ValueError
            If version and branch name cannot be found in the output string.
        RuntimeError
            If any unhandled event during running the subprocess occurs.
        """
        warnings.warn(
            "This method will be deprecated in a future release."
            "Use the `version` attribute instead.",
            FutureWarning,
        )
        try:
            result = subprocess.run(
                [self.get_new_cadet_instance().cadet_cli_path, "--version"],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            version_output = result.stdout.strip()

            version_match = re.search(
                r"cadet-cli version ([\d.]+) \(([^)]+)\)", version_output
            )

            if version_match:
                cadet_version = version_match.group(1)
                branch_name = version_match.group(2)
                return cadet_version, branch_name
            else:
                raise ValueError("CADET version or branch name missing from output.")
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Command execution failed: {e}")

    @property
    def version(self) -> str:
        """str: The version of the Cadet installation."""
        cadet = self.get_new_cadet_instance()
        # Check for CADET-Python >= v1.1, which introduced the version property.
        # If it's not present, assume CADET-Python <= 1.0.4 and directly access the
        # CadetRunner attribute.
        # This check can be removed at some point in the future.
        if hasattr(cadet, "version"):
            return cadet.version
        else:
            return cadet.cadet_runner.cadet_version

    def get_new_cadet_instance(self) -> CadetAPI:
        """Return new CadetAPI instance."""
        cadet = CadetAPI(install_path=self.install_path, use_dll=self.use_dll)
        return cadet

    def save_to_h5(self, process: Process, file_path: os.PathLike) -> None:
        """Save process config as h5 file."""
        cadet = self.get_new_cadet_instance()
        cadet.root = self.get_process_config(process)
        cadet.filename = file_path
        cadet.save()

    def run_h5(self, file_path: os.PathLike) -> CadetAPI:
        """Load and run h5 file."""
        cadet = self.get_new_cadet_instance()
        cadet.filename = file_path
        cadet.load()

        # Check for CADET-Python >= v1.1, which introduced the .run_simulation interface.
        # If it's not present, assume CADET-Python <= 1.0.4 and use the old .run_load() interface
        # This check can be removed at some point in the future.
        if hasattr(cadet, "run_simulation"):
            return_information = cadet.run_simulation(timeout=self.timeout)
        else:
            return_information = cadet.run_load(timeout=self.timeout)

        if return_information.return_code != 0:
            raise CADETProcessError(return_information.error_message)

        return cadet

    def load_from_h5(self, file_path: os.PathLike) -> CadetAPI:
        """Load data from HDF5 file."""
        cadet = self.get_new_cadet_instance()
        cadet.filename = file_path
        cadet.load()

        return cadet

    @wraps(CadetAPI.create_lwe)
    def create_lwe(self, *args: Any, **kwargs: Any) -> CadetAPI:
        """Wrap CadetAPI.create_lwe to create simple Load-Wash-Elute config."""
        return self.get_new_cadet_instance().create_lwe(*args, **kwargs)

    @locks_process
    def get_process_config(self, process: Process) -> Dict:
        """
        Create the CADET config.

        Returns
        -------
        config : Dict
            /

        Notes
        -----
            Sensitivities not implemented yet.

        See Also
        --------
        get_input_model
        get_input_solver
        get_input_return
        get_input_sensitivity
        """
        config = Dict()
        config.input.model = self.get_input_model(process)
        config.input.solver = self.get_input_solver(process)
        config.input["return"] = self.get_input_return(process)
        config.input.sensitivity = self.get_input_sensitivity(process)

        return config

    def load_simulation_results(
        self,
        process: Process,
        file_path: os.PathLike,
    ) -> SimulationResults:
        """Load simulation results."""
        cadet = self.load_from_h5(file_path)
        results = self.get_simulation_results(process, cadet)

        return results

    @locks_process
    def get_simulation_results(
        self,
        process: Process,
        cadet: CadetAPI,
        time_elapsed: Optional[float] = None,
        return_information: Optional[dict] = None,
    ) -> SimulationResults:
        """
        Read simulation results from CADET configuration.

        Parameters
        ----------
        process :  Process
            Process that was simulated.
        cadet : CadetAPI
            Cadet object with simulation results.
        time_elapsed : float
            Time of simulation.
        return_information: ReturnInformation
            CADET-cli return information.

        Returns
        -------
        results : SimulationResults
            Simulation results including process and solver configuration.

        ..todo::
            Implement method to read .h5 files that have no process associated.
        """
        if time_elapsed is None:
            time_elapsed = cadet.root.meta.time_sim
        time = self.get_solution_time(process)
        if return_information is None:
            exit_flag = None
            exit_message = None
        else:
            exit_flag = return_information.return_code
            exit_message = return_information.error_message

        try:
            solution = Dict()
            for unit in process.flow_sheet.units:
                solution[unit.name] = defaultdict(list)

                port_flag = unit.has_ports

                if port_flag:
                    solution[unit.name]["inlet"] = defaultdict(list)
                    solution[unit.name]["outlet"] = defaultdict(list)

                unit_index = self.get_unit_index(process, unit)
                unit_solution = cadet.root.output.solution[unit_index]

                unit_coordinates = cadet.root.output.coordinates[unit_index].copy()
                particle_coordinates = unit_coordinates.pop('particle_coordinates_000', None)

                for port in unit.ports:
                    port_index = self.get_port_index(process.flow_sheet, unit, port)

                    flow_in = process.flow_rate_timelines[unit.name].total_in[port]
                    flow_out = process.flow_rate_timelines[unit.name].total_out[port]

                    start = 0
                    for cycle in range(self.n_cycles):
                        end = start + len(time)

                        if (
                            f"solution_inlet_port_{port_index:03d}"
                            in unit_solution.keys()
                        ):
                            sol_inlet = unit_solution[
                                f"solution_inlet_port_{port_index:03d}"
                            ][start:end,]
                            if port_flag:
                                solution[unit.name]["inlet"][port].append(
                                    SolutionIO(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sol_inlet,
                                        flow_in,
                                    )
                                )
                            else:
                                solution[unit.name]["inlet"].append(
                                    SolutionIO(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sol_inlet,
                                        flow_in,
                                    )
                                )

                        if (
                            f"solution_outlet_port_{port_index:03d}"
                            in unit_solution.keys()
                        ):
                            sol_outlet = unit_solution[
                                f"solution_outlet_port_{port_index:03d}"
                            ][start:end, :]
                            if port_flag:
                                solution[unit.name]["outlet"][port].append(
                                    SolutionIO(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sol_outlet,
                                        flow_out,
                                    )
                                )
                            else:
                                solution[unit.name]["outlet"].append(
                                    SolutionIO(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sol_outlet,
                                        flow_out,
                                    )
                                )

                        if "solution_bulk" in unit_solution.keys():
                            sol_bulk = unit_solution.solution_bulk[start:end, :]
                            solution[unit.name]["bulk"].append(
                                SolutionBulk(
                                    unit.name,
                                    unit.component_system,
                                    time,
                                    sol_bulk,
                                    **unit_coordinates,
                                )
                            )

                        if "solution_particle" in unit_solution.keys():
                            sol_particle = unit_solution.solution_particle[start:end, :]
                            solution[unit.name]["particle"].append(
                                SolutionParticle(
                                    unit.name,
                                    unit.component_system,
                                    time,
                                    sol_particle,
                                    **unit_coordinates,
                                    particle_coordinates=particle_coordinates,
                                )
                            )

                        if "solution_solid" in unit_solution.keys():
                            sol_solid = unit_solution.solution_solid[start:end, :]
                            solution[unit.name]["solid"].append(
                                SolutionSolid(
                                    unit.name,
                                    unit.component_system,
                                    unit.binding_model.bound_states,
                                    time,
                                    sol_solid,
                                    **unit_coordinates,
                                    particle_coordinates=particle_coordinates,
                                )
                            )

                        if "solution_volume" in unit_solution.keys():
                            sol_volume = unit_solution.solution_volume[start:end]
                            solution[unit.name]["volume"].append(
                                SolutionVolume(
                                    unit.name, unit.component_system, time, sol_volume
                                )
                            )
                        start = end - 1

            solution = Dict(solution)

            sensitivity = Dict()
            for i, sens in enumerate(process.parameter_sensitivities):
                sens_index = f"param_{i:03d}"

                for unit in process.flow_sheet.units:
                    sensitivity[sens.name][unit.name] = defaultdict(list)

                    port_flag = unit.has_ports

                    if port_flag:
                        sensitivity[sens.name][unit.name]["inlet"] = defaultdict(list)
                        sensitivity[sens.name][unit.name]["outlet"] = defaultdict(list)

                    unit_index = self.get_unit_index(process, unit)
                    unit_sensitivity = \
                        cadet.root.output.sensitivity[sens_index][unit_index]
                    unit_coordinates = \
                        cadet.root.output.coordinates[unit_index].copy()
                    particle_coordinates = \
                        unit_coordinates.pop("particle_coordinates_000", None)

                    for port in unit.ports:
                        port_index = self.get_port_index(process.flow_sheet, unit, port)

                        flow_in = process.flow_rate_timelines[unit.name].total_in[port]
                        flow_out = \
                            process.flow_rate_timelines[unit.name].total_out[port]

                        start = 0
                        for cycle in range(self.n_cycles):
                            end = start + len(time)

                            if (
                                f"sens_inlet_port_{port_index:03d}"
                                in unit_sensitivity.keys()
                            ):
                                sens_inlet = unit_sensitivity[
                                    f"sens_inlet_port_{port_index:03d}"
                                ][start:end, :]
                                if port_flag:
                                    sensitivity[sens.name][unit.name]["inlet"][
                                        port
                                    ].append(
                                        SolutionIO(
                                            unit.name,
                                            unit.component_system,
                                            time,
                                            sens_inlet,
                                            flow_in,
                                        )
                                    )

                                else:
                                    sensitivity[sens.name][unit.name]["inlet"].append(
                                        SolutionIO(
                                            unit.name,
                                            unit.component_system,
                                            time,
                                            sens_inlet,
                                            flow_in,
                                        )
                                    )

                            if (
                                f"sens_outlet_port_{port_index:03d}"
                                in unit_sensitivity.keys()
                            ):
                                sens_outlet = \
                                    unit_sensitivity[f"sens_outlet_port_{port_index:03d}"][start:end, :]  # noqa: E501
                                if port_flag:
                                    sensitivity[sens.name][unit.name]["outlet"][
                                        port
                                    ].append(
                                        SolutionIO(
                                            unit.name,
                                            unit.component_system,
                                            time,
                                            sens_outlet,
                                            flow_out,
                                        )
                                    )
                                else:
                                    sensitivity[sens.name][unit.name]["outlet"].append(
                                        SolutionIO(
                                            unit.name,
                                            unit.component_system,
                                            time,
                                            sens_outlet,
                                            flow_out,
                                        )
                                    )

                            if "sens_bulk" in unit_sensitivity.keys():
                                sens_bulk = unit_sensitivity.sens_bulk[start:end, :]
                                sensitivity[sens.name][unit.name]["bulk"].append(
                                    SolutionBulk(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sens_bulk,
                                        **unit_coordinates,
                                    )
                                )

                            if "sens_particle" in unit_sensitivity.keys():
                                sens_particle = \
                                    unit_sensitivity.sens_particle[start:end, :]
                                sensitivity[sens.name][unit.name]["particle"].append(
                                    SolutionParticle(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sens_particle,
                                        **unit_coordinates,
                                        particle_coordinates=particle_coordinates,
                                    )
                                )

                            if "sens_solid" in unit_sensitivity.keys():
                                sens_solid = unit_sensitivity.sens_solid[start:end, :]
                                sensitivity[sens.name][unit.name]["solid"].append(
                                    SolutionSolid(
                                        unit.name,
                                        unit.component_system,
                                        unit.binding_model.bound_states,
                                        time,
                                        sens_solid,
                                        **unit_coordinates,
                                        particle_coordinates=particle_coordinates,
                                    )
                                )

                            if "sens_volume" in unit_sensitivity.keys():
                                sens_volume = unit_sensitivity.sens_volume[start:end, :]
                                sensitivity[sens.name][unit.name]["volume"].append(
                                    SolutionVolume(
                                        unit.name,
                                        unit.component_system,
                                        time,
                                        sens_volume,
                                    )
                                )

                            start = end - 1

            sensitivity = Dict(sensitivity)

            system_state = {
                "state": cadet.root.output.last_state_y,
                "state_derivative": cadet.root.output.last_state_ydot,
            }

            chromatograms = [
                solution[chrom.name].outlet[-1]
                for chrom in process.flow_sheet.product_outlets
            ]

        except KeyError:
            raise CADETProcessError("Results don't match Process")

        results = SimulationResults(
            solver_name=str(self),
            solver_parameters=dict(),
            exit_flag=exit_flag,
            exit_message=exit_message,
            time_elapsed=time_elapsed,
            process=process,
            solution_cycles=solution,
            sensitivity_cycles=sensitivity,
            system_state=system_state,
            chromatograms=chromatograms,
        )

        return results

    def get_input_model(self, process: Process) -> Dict:
        """
        Config branch `/input/model/`.

        Notes
        -----
        !!! External functions not implemented yet

        See Also
        --------
        model_connections
        model_solver
        model_units
        input_model_parameters
        """
        input_model = Dict()

        input_model.connections = self.get_model_connections(process)
        # input_model.external = self.model_external # !!! not working yet
        input_model.solver = self.model_solver_parameters.parameters
        input_model.update(self.get_model_units(process))

        if process.system_state is not None:
            input_model["INIT_STATE_Y"] = process.system_state
        if process.system_state_derivative is not None:
            input_model["INIT_STATE_YDOT"] = process.system_state_derivative

        return input_model

    def get_model_connections(self, process: Process) -> Dict:
        """Config branch `/input/model/connections`."""
        model_connections = Dict()
        if self._force_constant_flow_rate:
            model_connections["CONNECTIONS_INCLUDE_DYNAMIC_FLOW"] = 0
        else:
            model_connections["CONNECTIONS_INCLUDE_DYNAMIC_FLOW"] = 1

        model_connections["CONNECTIONS_INCLUDE_PORTS"] = 1

        index = 0

        section_states = process.flow_rate_section_states

        for cycle in range(0, self.n_cycles):
            for flow_rates_state in section_states.values():
                switch_index = f"switch_{index:03d}"
                model_connections[switch_index].section = index

                connections = self.cadet_connections(
                    flow_rates_state, process.flow_sheet
                )
                model_connections[switch_index].connections = connections
                index += 1

        model_connections.nswitches = index

        return model_connections

    def cadet_connections(self, flow_rates: dict, flow_sheet: FlowSheet) -> list:
        """list: Connections matrix for flow_rates state.

        Parameters
        ----------
        flow_rates : dict
            UnitOperations with outgoing flow rates.
        flow_sheet : FlowSheet
            Object which hosts units (for getting unit index).

        Returns
        -------
        ls : list
            Connections matrix for DESCRIPTION.

        """
        table = Dict()
        enum = 0

        for origin, unit_flow_rates in flow_rates.items():
            origin = flow_sheet[origin]
            origin_index = flow_sheet.get_unit_index(origin)
            for origin_port, dest_dict in unit_flow_rates.destinations.items():
                for dest in dest_dict:
                    for dest_port, flow_rate in dest_dict[dest].items():
                        destination = flow_sheet[dest]
                        destination_index = flow_sheet.get_unit_index(destination)
                        if np.any(flow_rate):
                            origin_port_red = flow_sheet.get_port_index(
                                origin, origin_port
                            )
                            dest_port_red = flow_sheet.get_port_index(
                                destination, dest_port
                            )

                            table[enum] = []
                            table[enum].append(int(origin_index))
                            table[enum].append(int(destination_index))
                            table[enum].append(int(origin_port_red))
                            table[enum].append(int(dest_port_red))
                            table[enum].append(-1)
                            table[enum].append(-1)
                            Q = flow_rate.tolist()
                            if self._force_constant_flow_rate:
                                table[enum] += [Q[0]]
                            else:
                                table[enum] += Q
                            enum += 1

        ls = []
        for connection in table.values():
            ls += connection

        return ls

    def get_unit_index(self, process: Process, unit: UnitBaseClass) -> str:
        """
        Return unit index in CADET format unit_xxx.

        Parameters
        ----------
        process : Process
            process to be simulated
        unit : UnitOperation
            Indexed object

        Returns
        -------
        unit_index : str
            Return the unit index in CADET format unit_XXX
        """
        index = process.flow_sheet.get_unit_index(unit)
        return f"unit_{index:03d}"

    def get_port_index(self, flow_sheet: FlowSheet, unit: UnitBaseClass, port: str) -> int:
        """
        Return port index in CADET format xxx.

        Parameters
        ----------
        port : string
            Indexed port
        unit : UnitOperation
            port of unit

        Returns
        -------
        port_index : index
            Return the port_index in CADET format xxx
        """
        return flow_sheet.get_port_index(unit, port)

    def get_model_units(self, process: Process) -> Dict:
        """
        Config branches for all units `/input/model/unit_000 ... unit_xxx`.

        See Also
        --------
        get_unit_config
        get_unit_index
        """
        model_units = Dict()

        model_units.nunits = len(process.flow_sheet.units)

        for unit in process.flow_sheet.units:
            unit_index = self.get_unit_index(process, unit)
            model_units[unit_index] = self.get_unit_config(unit)

        self.set_section_dependent_parameters(model_units, process)

        return model_units

    def get_unit_config(self, unit: UnitBaseClass) -> Dict:
        """
        Config branch `/input/model/unit_xxx` for individual unit.

        The unit operation parameters are converted to CADET format

        Notes
        -----
        In CADET, the parameter unit_config['discretization'].NBOUND should be
        moved to binding config or unit config

        See Also
        --------
        get_adsorption_config
        """
        unit_parameters = UnitParameters(unit)

        unit_config = Dict(unit_parameters.parameters)

        if not isinstance(unit.binding_model, NoBinding):
            if unit.binding_model.n_binding_sites > 1:
                n_bound = \
                    [unit.binding_model.n_binding_sites] * unit.binding_model.n_comp
            else:
                n_bound = unit.binding_model.bound_states

            unit_config["adsorption"] = self.get_adsorption_config(unit.binding_model)
            unit_config["adsorption_model"] = \
                unit_config["adsorption"]["ADSORPTION_MODEL"]
        else:
            n_bound = unit.n_comp * [0]

        if not isinstance(unit.discretization, NoDiscretization):
            unit_config["discretization"] = unit.discretization.parameters
            unit_config["discretization"]["spatial_method"] = \
                unit.discretization.spatial_method

        if isinstance(unit, Cstr) and not isinstance(unit.binding_model, NoBinding):
            unit_config["nbound"] = n_bound
        else:
            unit_config["discretization"]["nbound"] = n_bound

        # Bulk Reaction
        if not isinstance(unit.bulk_reaction_model, NoReaction):
            parameters = self.get_reaction_config(unit.bulk_reaction_model)

            unit_config["reaction_model"] = parameters["REACTION_MODEL"]
            # Converting bulk reaction to particle reaction interface (used by LRM)
            if isinstance(unit, TubularReactor):
                for key, value in parameters.items():
                    key = key.replace("bulk", "liquid")
                    unit_config["reaction"][key] = value
            else:
                unit_config["reaction_bulk"] = parameters

        # Particle Reaction
        if not isinstance(unit.particle_reaction_model, NoReaction):
            parameters = self.get_reaction_config(unit.particle_reaction_model)
            if isinstance(unit, LumpedRateModelWithoutPores):
                unit_config["reaction_model"] = parameters["REACTION_MODEL"]
                unit_config["reaction"] = parameters
            else:
                unit_config["reaction_model_particles"] = parameters["REACTION_MODEL"]
                unit_config["reaction_particle"].update(parameters)

        if isinstance(unit, Inlet):
            unit_config["sec_000"]["const_coeff"] = unit.c[:, 0]
            unit_config["sec_000"]["lin_coeff"] = unit.c[:, 1]
            unit_config["sec_000"]["quad_coeff"] = unit.c[:, 2]
            unit_config["sec_000"]["cube_coeff"] = unit.c[:, 3]

        return unit_config

    def set_section_dependent_parameters(self, model_units: Dict, process: Process) -> None:
        """Add time dependent model parameters to units."""
        section_states = process.section_states.values()

        section_index = 0
        for cycle in range(0, self.n_cycles):
            for param_states in section_states:
                for param, state in param_states.items():
                    param = param.split(".")
                    unit_name = param[1]
                    param_name = param[-1]
                    try:
                        unit = process.flow_sheet[unit_name]
                    except KeyError:
                        if unit_name == "output_states":
                            continue
                        else:
                            raise CADETProcessError(
                                "Unexpected section dependent parameter"
                            )
                    if param_name == "flow_rate":
                        continue
                    unit_index = process.flow_sheet.get_unit_index(unit)
                    if isinstance(unit, Inlet) and param_name == "c":
                        self.add_inlet_section(
                            model_units, section_index, unit_index, state
                        )
                    else:
                        unit_model = unit.model
                        self.add_parameter_section(
                            model_units,
                            section_index,
                            unit_index,
                            unit_model,
                            param_name,
                            state,
                        )

                section_index += 1

    def add_inlet_section(
        self, model_units: Dict, sec_index: int, unit_index: int, coeffs: np.ndarray
    ) -> None:
        """Add piecewise cubic polynomial section to inlet."""
        unit_index = f"unit_{unit_index:03d}"
        section_index = f"sec_{sec_index:03d}"

        model_units[unit_index][section_index]["const_coeff"] = coeffs[:, 0]
        model_units[unit_index][section_index]["lin_coeff"] = coeffs[:, 1]
        model_units[unit_index][section_index]["quad_coeff"] = coeffs[:, 2]
        model_units[unit_index][section_index]["cube_coeff"] = coeffs[:, 3]

    def add_parameter_section(
        self,
        model_units: Dict,
        sec_index: int,
        unit_index: int,
        unit_model: UnitBaseClass,
        parameter: str,
        state: np.ndarray,
    ) -> None:
        """Add section value to parameter branch."""
        unit_index = f"unit_{unit_index:03d}"
        parameter_name = inv_unit_parameters_map[unit_model]["parameters"][parameter]

        if sec_index == 0:
            model_units[unit_index][parameter_name] = []
        model_units[unit_index][parameter_name] += list(state.ravel())

    def get_adsorption_config(self, binding: BindingBaseClass) -> Dict:
        """
        Config branch `/input/model/unit_xxx/adsorption` for individual unit.

        Binding model parameters are extracted and converted to CADET format.

        Parameters
        ----------
        binding : BindingBaseClass
            Binding model

        See Also
        --------
        get_unit_config
        """
        adsorption_config = AdsorptionParameters(binding).parameters

        return adsorption_config

    def get_reaction_config(self, reaction: ReactionBaseClass) -> Dict:
        """
        Config branch `/input/model/unit_xxx/reaction` for individual unit.

        Reaction model parameters are extracted and converted to CADET format.

        Parameters
        ----------
        reaction : ReactionBaseClass
            Reaction model

        See Also
        --------
        get_unit_config
        """
        reaction_config = ReactionParameters(reaction).parameters

        return reaction_config

    def get_input_solver(self, process: Process) -> Dict:
        """
        Config branch `/input/solver/`.

        See Also
        --------
        solver_sections
        solver_time_integrator
        """
        input_solver = Dict()

        input_solver.update(self.solver_parameters.parameters)
        input_solver.user_solution_times = self.get_solution_time_complete(process)
        input_solver.sections = self.get_solver_sections(process)
        input_solver.time_integrator = self.time_integrator_parameters.parameters

        return input_solver

    def get_solver_sections(self, process: Process) -> Dict:
        """Config branch `/input/solver/sections`."""
        solver_sections = Dict()

        solver_sections.nsec = self.n_cycles * process.n_sections

        solver_sections.section_times = self.get_section_times_complete(process)

        solver_sections.section_continuity = [0] * (solver_sections.nsec - 1)

        return solver_sections

    def get_input_return(self, process: Process) -> dict:
        """Config branch `/input/return`."""
        return_parameters = self.return_parameters.parameters
        unit_return_parameters = self.get_unit_return_parameters(process)
        return {**return_parameters, **unit_return_parameters}

    def get_unit_return_parameters(self, process: Process) -> Dict:
        """Config branches for all units `/input/return/unit_000 ... unit_xxx`."""
        unit_return_parameters = Dict()
        for unit in process.flow_sheet.units:
            unit_index = self.get_unit_index(process, unit)
            unit_return_parameters[unit_index] = unit.solution_recorder.parameters

        return unit_return_parameters

    def get_input_sensitivity(self, process: Process) -> dict:
        """Config branch `/input/sensitivity`."""
        sensitivity_parameters = self.sensitivity_parameters.parameters
        parameter_sensitivities = self.get_parameter_sensitivities(process)
        return {**sensitivity_parameters, **parameter_sensitivities}

    def get_parameter_sensitivities(self, process: Process) -> Dict:
        """Config branches for all parameter sensitivities `/input/sensitivity/param_000 ... param_xxx`."""  # noqa: E501
        parameter_sensitivities = Dict()
        parameter_sensitivities.nsens = process.n_sensitivities
        for i, sens in enumerate(process.parameter_sensitivities):
            sens_index = f"param_{i:03d}"
            parameter_sensitivities[sens_index] = \
                self.get_sensitivity_config(process, sens)

        return parameter_sensitivities

    def get_sensitivity_config(self, process: Process, sens: Dict) -> Dict:
        """Return sensitivity config."""
        config = Dict()

        unit_indices = []
        parameters = []
        components = []

        for param, unit, associated_model, comp, coeff in zip(
            sens.parameters,
            sens.units,
            sens.associated_models,
            sens.components,
            sens.polynomial_coefficients,
        ):
            unit_index = process.flow_sheet.get_unit_index(unit)
            unit_indices.append(unit_index)

            if associated_model is None:
                model = unit.model
                if model == "Inlet" and param == "c":
                    if coeff == 0:
                        coeff = "CONST_COEFF"
                    elif coeff == 1:
                        coeff = "CONST_COEFF"
                    elif coeff == 2:
                        coeff = "QUAD_COEFF"
                    elif coeff == 3:
                        coeff = "CUBE_COEFF"
                    parameter = coeff
                else:
                    parameter = inv_unit_parameters_map[model]["parameters"][param]
            else:
                model = associated_model.model
                if isinstance(associated_model, BindingBaseClass):
                    parameter = inv_adsorption_parameters_map[model]["parameters"][
                        param
                    ]
                if isinstance(associated_model, ReactionBaseClass):
                    parameter = inv_reaction_parameters_map[model]["parameters"][param]
            parameters.append(parameter)

            component_system = unit.component_system
            comp = -1 if comp is None else component_system.indices[comp]
            components.append(comp)

        config.sens_unit = unit_indices
        config.sens_name = parameters
        config.sens_comp = components

        config.sens_partype = -1  # !!! Check when multiple particle types enabled.
        if not all([index is None for index in sens.bound_state_indices]):
            config.sens_reaction = [
                -1 if index is None else index for index in sens.bound_state_indices
            ]
        else:
            config.sens_reaction = -1

        if not all([index is None for index in sens.bound_state_indices]):
            config.sens_boundphase = [
                -1 if index is None else index for index in sens.bound_state_indices
            ]
        else:
            config.sens_boundphase = -1

        if not all([index is None for index in sens.section_indices]):
            config.sens_section = [
                -1 if index is None else index for index in sens.section_indices
            ]
        else:
            config.sens_section = -1

        if not all([index is None for index in sens.abstols]):
            config.sens_abstol = sens.abstols

        config.factors = sens.factors

        return config

    def __str__(self) -> str:
        """Return string representation."""
        return "CADET"


class ModelSolverParameters(Structure):
    """
    Converter for model solver parameters from CADETProcess to CADET.

    Attributes
    ----------
    gs_type : {1, 0}, optional
        Valid modes:
        - 0: Classical Gram-Schmidet orthogonalization.
        - 1: Modified Gram-Schmidt.
        The default is 1.
    max_krylov : int, optional
        Size of the Krylov subspace in the iterative linear GMRES solver.
        The default is 0.
    max_restarts : int, optional
        Maximum number of restarts in the GMRES algorithm. If lack of memory is not an
        issue, better use a larger Krylov space than restarts.
        The default is 10.
    schur_safety : float, optional
        Schur safety factor.
        Influences the tradeoff between linear iterations and nonlinear error control
        The default is 1e-8.
    linear_solution_mode : int
        Valid modes:
        - 0: Automatically chose mode based on heuristic.
        - 1: Solve system of models in parallel
        - 2: Solve system of models sequentially (only for systems without cyclic connections)
        The default is 0.

    See Also
    --------
    Structure
    """

    gs_type = Switch(default=1, valid=[0, 1])
    max_krylov = UnsignedInteger(default=0)
    max_restarts = UnsignedInteger(default=10)
    schur_safety = UnsignedFloat(default=1e-8)
    linear_solution_mode = UnsignedInteger(default=0, ub=2)

    _parameters = [
        "gs_type",
        "max_krylov",
        "max_restarts",
        "schur_safety",
        "linear_solution_mode",
    ]


unit_parameters_map = {
    "GeneralRateModel": {
        "name": "GENERAL_RATE_MODEL",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "INIT_Q": "q",
            "INIT_CP": "cp",
            "COL_DISPERSION": "axial_dispersion",
            "COL_LENGTH": "length",
            "COL_POROSITY": "bed_porosity",
            "FILM_DIFFUSION": "film_diffusion",
            "PAR_POROSITY": "particle_porosity",
            "PAR_RADIUS": "particle_radius",
            "PORE_ACCESSIBILITY": "pore_accessibility",
            "PAR_DIFFUSION": "pore_diffusion",
            "PAR_SURFDIFFUSION": "surface_diffusion",
            "CROSS_SECTION_AREA": "cross_section_area",
            "VELOCITY": "flow_direction",
        },
        "fixed": {
            "COL_DISPERSION_MULTIPLEX": 3,
            "FILM_DIFFUSION_MULTIPLEX": 3,
            "PAR_DIFFUSION_MULTIPLEX": 3,
            "PAR_SURFDIFFUSION_MULTIPLEX": 3,
            "PORE_ACCESSIBILITY_MULTIPLEX": 3,
        },
    },
    "LumpedRateModelWithPores": {
        "name": "LUMPED_RATE_MODEL_WITH_PORES",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "INIT_CP": "cp",
            "INIT_Q": "q",
            "COL_DISPERSION": "axial_dispersion",
            "COL_LENGTH": "length",
            "COL_POROSITY": "bed_porosity",
            "FILM_DIFFUSION": "film_diffusion",
            "PAR_POROSITY": "particle_porosity",
            "PAR_RADIUS": "particle_radius",
            "PORE_ACCESSIBILITY": "pore_accessibility",
            "CROSS_SECTION_AREA": "cross_section_area",
            "VELOCITY": "flow_direction",
        },
        "fixed": {
            "COL_DISPERSION_MULTIPLEX": 3,
            "FILM_DIFFUSION_MULTIPLEX": 3,
            "PORE_ACCESSIBILITY_MULTIPLEX": 3,
        },
    },
    "LumpedRateModelWithoutPores": {
        "name": "LUMPED_RATE_MODEL_WITHOUT_PORES",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "INIT_Q": "q",
            "COL_DISPERSION": "axial_dispersion",
            "COL_LENGTH": "length",
            "TOTAL_POROSITY": "total_porosity",
            "CROSS_SECTION_AREA": "cross_section_area",
            "VELOCITY": "flow_direction",
        },
        "fixed": {
            "COL_DISPERSION_MULTIPLEX": 3,
        },
    },
    "TubularReactor": {
        "name": "LUMPED_RATE_MODEL_WITHOUT_PORES",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "COL_DISPERSION": "axial_dispersion",
            "COL_LENGTH": "length",
            "CROSS_SECTION_AREA": "cross_section_area",
            "VELOCITY": "flow_direction",
        },
        "fixed": {
            "TOTAL_POROSITY": 1,
            "COL_DISPERSION_MULTIPLEX": 3,
        },
    },
    "Cstr": {
        "name": "CSTR",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "INIT_Q": "q",
            "FLOWRATE_FILTER": "flow_rate_filter",
            "CONST_SOLID_VOLUME": "const_solid_volume",
            "INIT_LIQUID_VOLUME": "init_liquid_volume",
        },
    },
    "MCT": {
        "name": "MULTI_CHANNEL_TRANSPORT",
        "parameters": {
            "NCOMP": "n_comp",
            "INIT_C": "c",
            "COL_DISPERSION": "axial_dispersion",
            "COL_LENGTH": "length",
            "NCHANNEL": "nchannel",
            "CHANNEL_CROSS_SECTION_AREAS": "channel_cross_section_areas",
            "EXCHANGE_MATRIX": "exchange_matrix",
            "VELOCITY": "flow_direction",
        },
        "fixed": {
            "COL_DISPERSION_MULTIPLEX": 7,
        },
    },
    "Inlet": {
        "name": "INLET",
        "parameters": {
            "NCOMP": "n_comp",
        },
        "fixed": {
            "INLET_TYPE": "PIECEWISE_CUBIC_POLY",
        },
    },
    "Outlet": {
        "name": "OUTLET",
        "parameters": {
            "NCOMP": "n_comp",
        },
    },
    "MixerSplitter": {
        "name": "CSTR",
        "parameters": {
            "NCOMP": "n_comp",
        },
        "fixed": {"INIT_VOLUME": 1e-9, "INIT_C": [0]},
    },
}

inv_unit_parameters_map = {
    unit: {
        "name": values["name"],
        "parameters": {v: k for k, v in values["parameters"].items()},
    }
    for unit, values in unit_parameters_map.items()
}


class UnitParameters(ParameterWrapper):
    """
    Converter for UnitOperation parameters from CADETProcess to CADET.

    See Also
    --------
    ParameterWrapper
    AdsorptionParameters
    ReactionParameters
    """

    _baseClass = UnitBaseClass

    _unit_parameters = unit_parameters_map

    _model_parameters = _unit_parameters
    _model_type = "UNIT_TYPE"


adsorption_parameters_map = {
    "NoBinding": {
        "name": "NONE",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
        },
    },
    "Linear": {
        "name": "LINEAR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "LIN_KA": "adsorption_rate",
            "LIN_KD": "desorption_rate",
        },
    },
    "Langmuir": {
        "name": "MULTI_COMPONENT_LANGMUIR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCL_KA": "adsorption_rate",
            "MCL_KD": "desorption_rate",
            "MCL_QMAX": "capacity",
        },
    },
    "LangmuirLDF": {
        "name": "MULTI_COMPONENT_LANGMUIR_LDF",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCLLDF_KEQ": "equilibrium_constant",
            "MCLLDF_KKIN": "driving_force_coefficient",
            "MCLLDF_QMAX": "capacity",
        },
    },
    "LangmuirLDFLiquidPhase": {
        "name": "MULTI_COMPONENT_LANGMUIR_LDF_LIQUID_PHASE",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCLLDFC_KEQ": "equilibrium_constant",
            "MCLLDFC_KKIN": "driving_force_coefficient",
            "MCLLDFC_QMAX": "capacity",
        },
    },
    "BiLangmuir": {
        "name": "MULTI_COMPONENT_BILANGMUIR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCBL_KA": "adsorption_rate",
            "MCBL_KD": "desorption_rate",
            "MCBL_QMAX": "capacity",
        },
    },
    "BiLangmuirLDF": {
        "name": "MULTI_COMPONENT_BILANGMUIR_LDF",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCBLLDF_KEQ": "equilibrium_constant",
            "MCBLLDF_KKIN": "driving_force_coefficient",
            "MCBLLDF_QMAX": "capacity",
        },
    },
    "FreundlichLDF": {
        "name": "FREUNDLICH_LDF",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "FLDF_KKIN": "driving_force_coefficient",
            "FLDF_KF": "freundlich_coefficient",
            "FLDF_N": "exponent",
        },
    },
    "StericMassAction": {
        "name": "STERIC_MASS_ACTION",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "SMA_KA": "adsorption_rate",
            "SMA_KD": "desorption_rate",
            "SMA_LAMBDA": "capacity",
            "SMA_NU": "characteristic_charge",
            "SMA_SIGMA": "steric_factor",
            "SMA_REFC0": "reference_liquid_phase_conc",
            "SMA_REFQ": "reference_solid_phase_conc",
        },
    },
    "AntiLangmuir": {
        "name": "MULTI_COMPONENT_ANTILANGMUIR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCAL_KA": "adsorption_rate",
            "MCAL_KD": "desorption_rate",
            "MCAL_QMAX": "capacity",
            "MCAL_ANTILANGMUIR": "antilangmuir",
        },
    },
    "Spreading": {
        "name": "MULTI_COMPONENT_SPREADING",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MCSPR_KA": "adsorption_rate",
            "MCSPR_KD": "desorption_rate",
            "MCSPR_QMAX": "capacity",
            "MCSPR_K12": "exchange_from_1_2",
            "MCSPR_K21": "exchange_from_2_1",
        },
    },
    "MobilePhaseModulator": {
        "name": "MOBILE_PHASE_MODULATOR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MPM_KA": "adsorption_rate",
            "MPM_KD": "desorption_rate",
            "MPM_QMAX": "capacity",
            "MPM_BETA": "ion_exchange_characteristic",
            "MPM_GAMMA": "hydrophobicity",
            "MPM_LINEAR_THRESHOLD": "linear_threshold",
        },
    },
    "ExtendedMobilePhaseModulator": {
        "name": "EXTENDED_MOBILE_PHASE_MODULATOR",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "EMPM_KA": "adsorption_rate",
            "EMPM_KD": "desorption_rate",
            "EMPM_QMAX": "capacity",
            "EMPM_BETA": "ion_exchange_characteristic",
            "EMPM_GAMMA": "hydrophobicity",
            "EMPM_COMP_MODE": "component_mode",
        },
    },
    "SelfAssociation": {
        "name": "SELF_ASSOCIATION",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "SAI_KA1": "adsorption_rate",
            "SAI_KA2": "adsorption_rate_dimerization",
            "SAI_KD": "desorption_rate",
            "SAI_NU": "characteristic_charge",
            "SAI_SIGMA": "steric_factor",
            "SAI_LAMBDA": "capacity",
            "SAI_REFC0": "reference_liquid_phase_conc",
            "SAI_REFQ": "reference_solid_phase_conc",
        },
    },
    "BiStericMassAction": {
        "name": "BI_STERIC_MASS_ACTION",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "BISMA_KA": "adsorption_rate",
            "BISMA_KD": "desorption_rate",
            "BISMA_LAMBDA": "capacity",
            "BISMA_NU": "characteristic_charge",
            "BISMA_SIGMA": "steric_factor",
            "BISMA_REFC0": "reference_liquid_phase_conc",
            "BISMA_REFQ": "reference_solid_phase_conc",
        },
    },
    "MultistateStericMassAction": {
        "name": "MULTISTATE_STERIC_MASS_ACTION",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "MSSMA_KA": "adsorption_rate",
            "MSSMA_KD": "desorption_rate",
            "MSSMA_LAMBDA": "capacity",
            "MSSMA_NU": "characteristic_charge",
            "MSSMA_SIGMA": "steric_factor",
            "MSSMA_RATES": "conversion_rate",
            "MSSMA_REFC0": "reference_liquid_phase_conc",
            "MSSMA_REFQ": "reference_solid_phase_conc",
        },
    },
    "SimplifiedMultistateStericMassAction": {
        "name": "MULTISTATE_STERIC_MASS_ACTION",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "SMSSMA_KA": "adsorption_rate",
            "SMSSMA_KD": "desorption_rate",
            "SMSSMA_NU_MIN": "characteristic_charge_first",
            "SMSSMA_NU_MAX": "characteristic_charge_last",
            "SMSSMA_NU_QUAD": "quadratic_modifiers_charge",
            "SMSSMA_SIGMA_MIN": "steric_factor_first",
            "SMSSMA_SIGMA_MAX": "steric_factor_last",
            "SMSSMA_SIGMA_QUAD": "quadratic_modifiers_steric",
            "SMSSMA_LAMBDA": "capacity",
            "SMSSMA_KWS": "exchange_from_weak_stronger",
            "SMSSMA_KWS_LIN": "linear_exchange_ws",
            "SMSSMA_KWS_QUAD": "quadratic_exchange_ws",
            "SMSSMA_KSW": "exchange_from_stronger_weak",
            "SMSSMA_KSW_LIN": "linear_exchange_sw",
            "SMSSMA_KSW_QUAD": "quadratic_exchange_sw",
            "SMSSMA_REFC0": "reference_liquid_phase_conc",
            "SMSSMA_REFQ": "reference_solid_phase_conc",
        },
    },
    "Saska": {
        "name": "SASKA",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "SASKA_H": "henry_const",
            "SASKA_K": "quadratic_factor",
        },
    },
    "GeneralizedIonExchange": {
        "name": "GENERALIZED_ION_EXCHANGE",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "GIEX_KA": "adsorption_rate",
            "GIEX_KA_LIN": "adsorption_rate_linear",
            "GIEX_KA_QUAD": "adsorption_rate_quadratic",
            "GIEX_KA_CUBE": "adsorption_rate_cubic",
            "GIEX_KA_SALT": "adsorption_rate_salt",
            "GIEX_KA_PROT": "adsorption_rate_protein",
            "GIEX_KD": "desorption_rate",
            "GIEX_KD_LIN": "desorption_rate_linear",
            "GIEX_KD_QUAD": "desorption_rate_quadratic",
            "GIEX_KD_CUBE": "desorption_rate_cubic",
            "GIEX_KD_SALT": "desorption_rate_salt",
            "GIEX_KD_PROT": "desorption_rate_protein",
            "GIEX_NU_BREAKS": "characteristic_charge_breaks",
            "GIEX_NU": "characteristic_charge",
            "GIEX_NU_LIN": "characteristic_charge_linear",
            "GIEX_NU_QUAD": "characteristic_charge_quadratic",
            "GIEX_NU_CUBE": "characteristic_charge_cubic",
            "GIEX_SIGMA": "steric_factor",
            "GIEX_LAMBDA": "capacity",
            "GIEX_REFC0": "reference_liquid_phase_conc",
            "GIEX_REFQ": "reference_solid_phase_conc",
        },
    },
    "HICConstantWaterActivity": {
        "name": "HIC_CONSTANT_WATER_ACTIVITY",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "HICCWA_KA": "adsorption_rate",
            "HICCWA_KD": "desorption_rate",
            "HICCWA_NU": "hic_characteristic",
            "HICCWA_QMAX": "capacity",
            "HICCWA_BETA0": "beta_0",
            "HICCWA_BETA1": "beta_1",
        },
    },
    "HICWaterOnHydrophobicSurfaces": {
        "name": "HIC_WATER_ON_HYDROPHOBIC_SURFACES",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "HICWHS_KA": "adsorption_rate",
            "HICWHS_KD": "desorption_rate",
            "HICWHS_NU": "hic_characteristic",
            "HICWHS_QMAX": "capacity",
            "HICWHS_BETA0": "beta_0",
            "HICWHS_BETA1": "beta_1",
        },
    },
    "MultiComponentColloidal": {
        "name": "MULTI_COMPONENT_COLLOIDAL",
        "parameters": {
            "IS_KINETIC": "is_kinetic",
            "COL_PHI": "phase_ratio",
            "COL_KAPPA_EXP": "kappa_exponential",
            "COL_KAPPA_FACT": "kappa_factor",
            "COL_KAPPA_CONST": "kappa_constant",
            "COL_CORDNUM": "coordination_number",
            "COL_LOGKEQ_PH_EXP": "logkeq_ph_exponent",
            "COL_LOGKEQ_SALT_POWEXP": "logkeq_power_exponent",
            "COL_LOGKEQ_SALT_POWFACT": "logkeq_power_factor",
            "COL_LOGKEQ_SALT_EXPFACT": "logkeq_exponent_factor",
            "COL_LOGKEQ_SALT_EXPARGMULT": "logkeq_exponent_multiplier",
            "COL_BPP_PH_EXP": "bpp_ph_exponent",
            "COL_BPP_SALT_POWEXP": "bpp_power_exponent",
            "COL_BPP_SALT_POWFACT": "bpp_power_factor",
            "COL_BPP_SALT_EXPFACT": "bpp_exponent_factor",
            "COL_BPP_SALT_EXPARGMULT": "bpp_exponent_multiplier",
            "COL_PROTEIN_RADIUS": "protein_radius",
            "COL_KKIN": "kinetic_rate_constant",
            "COL_LINEAR_THRESHOLD": "linear_threshold",
            "COL_USE_PH": "use_ph",
        },
    },
}

inv_adsorption_parameters_map = {
    model: {
        "name": values["name"],
        "parameters": {v: k for k, v in values["parameters"].items()},
    }
    for model, values in adsorption_parameters_map.items()
}


class AdsorptionParameters(ParameterWrapper):
    """
    Converter for Binding model parameters from CADETProcess to CADET.

    See Also
    --------
    ParameterWrapper
    ReactionParameters
    UnitParameters
    """

    _baseClass = BindingBaseClass

    _adsorption_parameters = adsorption_parameters_map

    _model_parameters = _adsorption_parameters
    _model_type = "ADSORPTION_MODEL"


reaction_parameters_map = {
    "NoReaction": {
        "name": "NONE",
        "parameters": {},
    },
    "MassActionLaw": {
        "name": "MASS_ACTION_LAW",
        "parameters": {
            "mal_stoichiometry_bulk": "stoich",
            "mal_exponents_bulk_fwd": "exponents_fwd",
            "mal_exponents_bulk_bwd": "exponents_bwd",
            "mal_kfwd_bulk": "k_fwd",
            "mal_kbwd_bulk": "k_bwd",
        },
    },
    "MassActionLawParticle": {
        "name": "MASS_ACTION_LAW",
        "parameters": {
            "mal_stoichiometry_liquid": "stoich_liquid",
            "mal_exponents_liquid_fwd": "exponents_fwd_liquid",
            "mal_exponents_liquid_bwd": "exponents_bwd_liquid",
            "mal_kfwd_liquid": "k_fwd_liquid",
            "mal_kbwd_liquid": "k_bwd_liquid",
            "mal_stoichiometry_solid": "stoich_solid",
            "mal_exponents_solid_fwd": "exponents_fwd_solid",
            "mal_exponents_solid_bwd": "exponents_bwd_solid",
            "mal_kfwd_solid": "k_fwd_solid",
            "mal_kbwd_solid": "k_bwd_solid",
            "mal_exponents_liquid_fwd_modsolid": "exponents_fwd_liquid_modsolid",
            "mal_exponents_liquid_bwd_modsolid": "exponents_bwd_liquid_modsolid",
            "mal_exponents_solid_fwd_modliquid": "exponents_fwd_solid_modliquid",
            "mal_exponents_solid_bwd_modliquid": "exponents_bwd_solid_modliquid",
        },
    },
}


inv_reaction_parameters_map = {
    model: {
        "name": values["name"],
        "parameters": {v: k for k, v in values["parameters"].items()},
    }
    for model, values in adsorption_parameters_map.items()
}


class ReactionParameters(ParameterWrapper):
    """
    Converter for Reaction model parameters from CADETProcess to CADET.

    See Also
    --------
    ParameterWrapper
    AdsorptionParameters
    UnitParameters
    """

    _baseClass = ReactionBaseClass

    _reaction_parameters = reaction_parameters_map

    _model_parameters = _reaction_parameters
    _model_type = "REACTION_MODEL"


class SolverParameters(Structure):
    """
    Class for defining the solver parameters for CADET.

    Attributes
    ----------
    nthreads : int
        Number of used threads.
        The default is 1.
    consistent_init_mode : int, optional
        Consistent initialization mode.
        Valid values are:
        - 0: None
        - 1: Full
        - 2: Once, full
        - 3: Lean
        - 4: Once, lean
        - 5: Full once, then lean
        - 6: None once, then full
        - 7: None once, then lean
        The default is 1.
    consistent_init_mode_sens : int, optional
        Consistent initialization mode for parameter sensitivities.
        Valid values are:
        - 0: None
        - 1: Full
        - 2: Once, full
        - 3: Lean
        - 4: Once, lean
        - 5: Full once, then lean
        - 6: None once, then full
        - 7: None once, then lean
        The default is 1.

    """

    nthreads = UnsignedInteger(default=1)
    consistent_init_mode = UnsignedInteger(default=1, ub=7)
    consistent_init_mode_sens = UnsignedInteger(default=1, ub=7)

    _parameters = ["nthreads", "consistent_init_mode", "consistent_init_mode_sens"]


class SolverTimeIntegratorParameters(Structure):
    """
    Converter for time integartor parameters from CADETProcess to CADET.

    Attributes
    ----------
    abstol: float, optional
        Absolute tolerance in the solution of the original system.
        The default is 1e-8.
    algtol: float, optional
        Tolerance in the solution of the nonlinear consistency equations.
        The default is 1e-12.
    reltol: float, optional
        Relative tolerance in the solution of the original system.
        The default is 1e-6.
    reltol_sens: float, optional
        Relative tolerance in the solution of the sensitivity systems.
        The default is 1e-12.
    init_step_size: float, optional
        Initial time integrator step size.
        The default is 1e-6.
    max_steps: int, optional
        Maximum number of timesteps taken by IDAS
        The default is 1000000.
    max_step_size: float, optional
        Maximum size of timesteps taken by IDAS.
        The default is 0.0 (unlimited).
    errortest_sens: bool, optional
        If True: Use (forward) sensitivities in local error test
        The default is True.
    max_newton_iter: int, optional
        Maximum number of Newton iterations in time step.
        The default is 3.
    max_errtest_fail: int, optional
        Maximum number of local error test failures in time step
        The default is 7.
    max_convtest_fail: int, optional
        Maximum number of Newton convergence test failures
        The default is 10.
    max_newton_iter_sens: int, optional
        Maximum number of Newton iterations in forward sensitivity time step
        The default is 3.

    See Also
    --------
    Structure
    """

    abstol = UnsignedFloat(default=1e-8)
    algtol = UnsignedFloat(default=1e-12)
    reltol = UnsignedFloat(default=1e-6)
    reltol_sens = UnsignedFloat(default=1e-12)
    init_step_size = UnsignedFloat(default=1e-6)
    max_steps = UnsignedInteger(default=1000000)
    max_step_size = UnsignedFloat(default=0.0)
    errortest_sens = Bool(default=False)
    max_newton_iter = UnsignedInteger(default=1000000)
    max_errtest_fail = UnsignedInteger(default=1000000)
    max_convtest_fail = UnsignedInteger(default=1000000)
    max_newton_iter_sens = UnsignedInteger(default=1000000)

    _parameters = [
        "abstol",
        "algtol",
        "reltol",
        "reltol_sens",
        "init_step_size",
        "max_steps",
        "max_step_size",
        "errortest_sens",
        "max_newton_iter",
        "max_errtest_fail",
        "max_convtest_fail",
        "max_newton_iter_sens",
    ]


class ReturnParameters(Structure):
    """Solution writer for system."""

    write_solution_times = Bool(default=True)
    write_solution_last = Bool(default=True)
    write_sens_last = Bool(default=True)
    split_components_data = Bool(default=False)
    split_ports_data = Bool(default=True)
    single_as_multi_port = Bool(default=True)

    _parameters = [
        "write_solution_times",
        "write_solution_last",
        "write_sens_last",
        "split_components_data",
        "split_ports_data",
        "single_as_multi_port",
    ]


class SensitivityParameters(Structure):
    """Sensitivity parameters."""

    sens_method = Switch(default="ad1", valid=["ad1"])

    _parameters = ["sens_method"]



from __future__ import annotations

from collections import defaultdict
from functools import wraps
from typing import Any, Callable, Iterator, Optional
from warnings import warn

import numpy as np
from addict import Dict

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import String, Structure, frozen_attributes

from .binding import NoBinding
from .componentSystem import ComponentSystem
from .unitOperation import Cstr, Inlet, Outlet, SourceMixin, UnitBaseClass


@frozen_attributes
class FlowSheet(Structure):
    """
    Class to design process flow sheet.

    In this class, UnitOperation models are added and connected in a flow
    sheet.

    Attributes
    ----------
    n_comp : UnsignedInteger
        Number of components of the units in the flow sheet.
    name : String
        Name of the FlowSheet.
    units : list
        UnitOperations in the FlowSheet.
    connections : dict
        Connections of UnitOperations.
    output_states : dict
        Split ratios of outgoing streams of UnitOperations.
    """

    name = String()

    def __init__(
        self,
        component_system: ComponentSystem,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize flow sheet."""
        super().__init__(*args, **kwargs)

        self.component_system = component_system

        self._units = []
        self._feed_inlets = []
        self._eluent_inlets = []
        self._product_outlets = []
        self._connections = Dict()
        self._output_states = Dict()
        self._flow_rates = Dict()
        self._parameters = Dict()
        self._sized_parameters = Dict()
        self._polynomial_parameters = Dict()
        self._section_dependent_parameters = Dict()

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: The component system of the flow sheet."""
        return self._component_system

    @component_system.setter
    def component_system(self, component_system: ComponentSystem) -> None:
        if not isinstance(component_system, ComponentSystem):
            raise TypeError("Expected ComponentSystem")
        self._component_system = component_system

    @property
    def n_comp(self) -> int:
        """int: The number of components."""
        return self.component_system.n_comp

    def unit_name_decorator(func: Callable) -> Callable:
        """Wrap methods to enable calling functions with unit object or unit name."""

        @wraps(func)
        def unit_name_wrapper(
            self: FlowSheet,
            unit: str | UnitBaseClass,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Enable calling functions with unit object or unit name."""
            if isinstance(unit, str):
                try:
                    unit = self.units_dict[unit]
                except KeyError:
                    raise CADETProcessError("Not a valid unit")
            return func(self, unit, *args, **kwargs)

        return unit_name_wrapper

    def origin_destination_name_decorator(func: Callable) -> Callable:
        """Wrap methods to enable calling functions using origin and destination units."""

        @wraps(func)
        def origin_destination_name_wrapper(
            self: FlowSheet,
            origin: str | UnitBaseClass,
            destination: str | UnitBaseClass,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Enable calling origin and destination using unit names."""
            if isinstance(origin, str):
                try:
                    origin = self.units_dict[origin]
                except KeyError:
                    raise CADETProcessError("Not a valid unit")

            if isinstance(destination, str):
                try:
                    destination = self.units_dict[destination]
                except KeyError:
                    raise CADETProcessError("Not a valid unit")

            return func(self, origin, destination, *args, **kwargs)

        return origin_destination_name_wrapper

    def update_parameters(self) -> None:
        """Update current parameters."""
        for unit in self.units:
            self._parameters[unit.name] = unit.parameters
            self._section_dependent_parameters[unit.name] = (
                unit.section_dependent_parameters
            )
            self._polynomial_parameters[unit.name] = unit.polynomial_parameters
            self._sized_parameters[unit.name] = unit.sized_parameters

        self._parameters["output_states"] = {
            unit.name: self.output_states[unit] for unit in self.units
        }

        self._sized_parameters["output_states"] = {
            unit.name: self.output_states[unit] for unit in self.units
        }

        self._section_dependent_parameters["output_states"] = {
            unit.name: self.output_states[unit] for unit in self.units
        }

    def update_parameters_decorator(func: Callable) -> Callable:
        """Wrap method s.t. parameters dict is automatically updated."""

        @wraps(func)
        def wrapper(self: FlowSheet, *args: Any, **kwargs: Any) -> Any:
            """Update parameters dict to save time."""
            results = func(self, *args, **kwargs)
            self.update_parameters()

            return results

        return wrapper

    @property
    def units(self) -> list[UnitBaseClass]:
        """list: list of all unit_operations in the flow sheet."""
        return self._units

    @property
    def units_dict(self) -> dict[str, UnitBaseClass]:
        """dict: Unit operation names and objects."""
        return {unit.name: unit for unit in self.units}

    @property
    def unit_names(self) -> list[str]:
        """list: Names of unit operations."""
        return [unit.name for unit in self.units]

    @property
    def number_of_units(self) -> int:
        """int: Number of unit operations in the FlowSheet."""
        return len(self._units)

    @unit_name_decorator
    def get_unit_index(self, unit: UnitBaseClass) -> int:
        """
        Return the unit index of the unit.

        Parameters
        ----------
        unit : UnitBaseClass
            UnitBaseClass object of which the index is to be returned.

        Raises
        ------
        CADETProcessError
            If unit does not exist in the current flow sheet.

        Returns
        -------
        unit_index : int
            Returns the unit index of the unit_operation.
        """
        if unit not in self.units:
            raise CADETProcessError("Unit not in flow sheet")

        return self.units.index(unit)

    def get_port_index(self, unit: UnitBaseClass, port: str) -> int:
        """
        Return the port index of a unit.

        Parameters
        ----------
        unit : UnitBaseClass
            UnitBaseClass object of wich port index is to be returned
        port : string
            Name of port which index is to be returned

        Raises
        ------
        CADETProcessError
            If unit or port is not in the current flow sheet.

        Returns
        -------
        port_index : int
            Returns the port index of the port of the unit_operation.
        """
        if unit not in self.units:
            raise CADETProcessError("Unit not in flow sheet")

        port_index = unit.ports.index(port)

        return port_index

    @property
    def inlets(self) -> list[Inlet]:
        """list: All Inlets in the system."""
        return [unit for unit in self._units if isinstance(unit, Inlet)]

    @property
    def outlets(self) -> list[Outlet]:
        """list: All Outlets in the system."""
        return [unit for unit in self._units if isinstance(unit, Outlet)]

    @property
    def cstrs(self) -> list[Cstr]:
        """list: All Cstrs in the system."""
        return [unit for unit in self._units if isinstance(unit, Cstr)]

    @property
    def units_with_binding(self) -> list[UnitBaseClass]:
        """list: UnitOperations with binding models."""
        return [
            unit
            for unit in self._units
            if not isinstance(unit.binding_model, NoBinding)
        ]

    @update_parameters_decorator
    def add_unit(
        self,
        unit: UnitBaseClass,
        feed_inlet: bool = False,
        eluent_inlet: bool = False,
        product_outlet: bool = False,
    ) -> None:
        """
        Add unit to the flow sheet.

        Parameters
        ----------
        unit : UnitBaseClass
            UnitBaseClass object to be added to the flow sheet.
        feed_inlet : bool
            If True, add unit to feed inlets.
        eluent_inlet : bool
            If True, add unit to eluent inlets.
        product_outlet : bool
            If True, add unit to product outlets.

        Raises
        ------
        TypeError
            If unit is no instance of UnitBaseClass.
        CADETProcessError
            If unit already exists in flow sheet.
            If n_comp does not match with FlowSheet.

        See Also
        --------
        remove_unit
        """
        if not isinstance(unit, UnitBaseClass):
            raise TypeError("Expected UnitOperation")

        if unit in self._units or unit.name in self.unit_names:
            raise CADETProcessError("Unit already part of System")

        if unit.component_system is not self.component_system:
            raise CADETProcessError("Component systems do not match.")

        self._units.append(unit)

        for port in unit.ports:
            if isinstance(unit, Inlet):
                self._connections[unit]["origins"] = None
                self._connections[unit]["destinations"][port] = defaultdict(list)

            elif isinstance(unit, Outlet):
                self._connections[unit]["origins"][port] = defaultdict(list)
                self._connections[unit]["destinations"] = None

            else:
                self._connections[unit]["origins"][port] = defaultdict(list)
                self._connections[unit]["destinations"][port] = defaultdict(list)

        self._output_states[unit] = Dict()
        self._flow_rates[unit] = []

        super().__setattr__(unit.name, unit)

        if feed_inlet:
            self.add_feed_inlet(unit)
        if eluent_inlet:
            self.add_eluent_inlet(unit)
        if product_outlet:
            self.add_product_outlet(unit)

    @unit_name_decorator
    @update_parameters_decorator
    def remove_unit(self, unit: UnitBaseClass) -> None:
        """
        Remove unit from flow sheet.

        Removes unit from the list. Tries to remove units which are twice
        located as desinations. For this the origins and destinations are
        deleted for the unit. Raises a CADETProcessError if an ValueError is
        excepted. If the unit is specified as feed_inlet, eluent_inlet
        or product_outlet, the corresponding attributes are deleted.

        Parameters
        ----------
        unit : UnitBaseClass
            UnitBaseClass object to be removed to the flow sheet.

        Raises
        ------
        CADETProcessError
            If unit does not exist in the flow sheet.

        See Also
        --------
        add_unit
        feed_inlet
        eluent_inlet
        product_outlet
        """
        if unit not in self.units:
            raise CADETProcessError("Unit not in flow sheet")

        if unit is self.feed_inlets:
            self.remove_feed_inlet(unit)
        if unit is self.eluent_inlets:
            self.remove_eluent_inlet(unit)
        if unit is self.product_outlets:
            self.remove_product_outlet(unit)

        origins = []
        destinations = []

        if self._connections[unit]["origins"] is not None:
            origins = [
                origin
                for ports in self._connections[unit]["origins"]
                for origin in self._connections[unit]["origins"][ports]
            ]

        if self._connections[unit]["destinations"] is not None:
            destinations = [
                destination
                for ports in self._connections[unit]["destinations"]
                for destination in self._connections[unit]["destinations"][ports]
            ].copy()

        for origin in origins:
            for origin_port in self._connections[unit]["origins"]:
                for unit_port in self._connections[unit]["origins"][origin_port][origin]:
                    self.remove_connection(origin, unit, origin_port, unit_port)

        for destination in destinations:
            for destination_port in self._connections[unit]["destinations"]:
                for unit_port in self._connections[unit]["destinations"][destination_port][destination]:  # noqa: E501
                    self.remove_connection(unit, destination, unit_port, destination_port)

        self._units.remove(unit)
        self._connections.pop(unit)
        self._output_states.pop(unit)
        self.__dict__.pop(unit.name)

    @property
    def connections(self) -> Dict:
        """dict: In- and outgoing connections for each unit.

        See Also
        --------
        add_connection
        remove_connection

        """
        return self._connections

    @origin_destination_name_decorator
    @update_parameters_decorator
    def add_connection(
        self,
        origin: UnitBaseClass,
        destination: UnitBaseClass,
        origin_port: Optional[str] = None,
        destination_port: Optional[str] = None,
    ) -> None:
        """
        Add connection between units 'origin' and 'destination'.

        Parameters
        ----------
        origin : UnitBaseClass
            UnitBaseClass from which the connection originates.
        destination : UnitBaseClass
            UnitBaseClass where the connection terminates.
        origin_port : str, optional
            Port from which connection originates.
        destination_port : str, optional
            Port where connection terminates.

        Raises
        ------
        CADETProcessError
            If origin OR destination do not exist in the current flow sheet.
            If connection already exists in the current flow sheet.

        See Also
        --------
        connections
        remove_connection
        output_state
        """
        if origin not in self._units:
            raise CADETProcessError("Origin not in flow sheet")
        if isinstance(origin, Outlet):
            raise CADETProcessError("Outlet unit cannot have outgoing stream.")

        if destination not in self._units:
            raise CADETProcessError("Destination not in flow sheet")
        if isinstance(destination, Inlet):
            raise CADETProcessError("Inlet unit cannot have ingoing stream.")

        if origin.has_ports and origin_port is None:
            raise CADETProcessError("Missing `origin_port`")
        if not origin.has_ports and origin_port is not None:
            raise CADETProcessError("Origin unit does not support ports.")
        if origin.has_ports and origin_port not in origin.ports:
            raise CADETProcessError(
                f'Origin port "{origin_port}" not found in ports: {origin.ports}.'
            )
        if (
            origin_port
            in self._connections[destination]["origins"][destination_port][origin]
        ):
            raise CADETProcessError("Connection already exists")

        if destination.has_ports and destination_port is None:
            raise CADETProcessError("Missing `destination_port`")
        if not destination.has_ports and destination_port is not None:
            raise CADETProcessError("Destination unit does not support ports.")
        if destination.has_ports and destination_port not in destination.ports:
            raise CADETProcessError(
                f'destination port "{destination_port}" not found in ports: {destination.ports}.'
            )

        if (
            destination_port
            in self._connections[origin]["destinations"][origin_port][destination]
        ):
            raise CADETProcessError("Connection already exists")

        if not destination.has_ports:
            destination_port = destination.ports[0]
        if not origin.has_ports:
            origin_port = origin.ports[0]

        self._connections[destination]["origins"][destination_port][origin].append(
            origin_port
        )
        self._connections[origin]["destinations"][origin_port][destination].append(
            destination_port
        )

        self.set_output_state(origin, 0, origin_port)

    @origin_destination_name_decorator
    @update_parameters_decorator
    def remove_connection(
        self,
        origin: UnitBaseClass,
        destination: UnitBaseClass,
        origin_port: Optional[str] = None,
        destination_port: Optional[str] = None,
    ) -> None:
        """
        Remove connection between units 'origin' and 'destination'.

        Parameters
        ----------
        origin : UnitBaseClass
            UnitBaseClass from which the connection originates.
        destination : UnitBaseClass
            UnitBaseClass where the connection terminates.
        origin_port : int
            Port from which connection originates.
        destination_port : int
            Port where connection terminates.

        Raises
        ------
        CADETProcessError
            If origin OR destination do not exist in the current flow sheet.
            If connection does not exists in the current flow sheet.

        See Also
        --------
        connections
        add_connection
        """
        if origin not in self._units:
            raise CADETProcessError("Origin not in flow sheet")
        if origin.has_ports and origin_port is None:
            raise CADETProcessError("Missing `origin_port`")

        if origin_port not in origin.ports:
            raise CADETProcessError(
                f"Origin port {origin_port} not in Unit {origin.name}."
            )

        if destination not in self._units:
            raise CADETProcessError("Destination not in flow sheet")
        if destination.has_ports and origin_port is None:
            raise CADETProcessError("Missing `destination_port`")

        if destination_port not in destination.ports:
            raise CADETProcessError(
                f"Origin port {destination_port} not in Unit {destination.name}."
            )

        try:
            self._connections[destination]["origins"][destination_port].pop(origin)
            self._connections[origin]["destinations"][origin_port].pop(destination)
        except KeyError:
            raise CADETProcessError("Connection does not exist.")

    @origin_destination_name_decorator
    def connection_exists(
        self,
        origin: UnitBaseClass,
        destination: UnitBaseClass,
        origin_port: Optional[str] = None,
        destination_port: Optional[str] = None,
    ) -> bool:
        """
        bool: check if connection exists in flow sheet.

        Parameters
        ----------
        origin : UnitBaseClass
            UnitBaseClass from which the connection originates.
        destination : UnitBaseClass
            UnitBaseClass where the connection terminates.
        origin_port : Port, optional
            If origin unit operation has ports, origin port can be specified.
        destination_port : Port optional
            if destination unit operation has ports, destination port can be specified.

        """
        if origin.has_ports and not origin_port:
            raise CADETProcessError(
                "Origin port needs to be specified for unit operation with ports "
                f"{origin.name}."
            )

        if destination.has_ports and not destination_port:
            raise CADETProcessError(
                "Destination port needs to be specified for unit operation with ports "
                f"{destination.name}."
            )

        if origin_port not in origin.ports:
            raise CADETProcessError(f"{origin.name} does not have port {origin_port}")

        if destination_port not in destination.ports:
            raise CADETProcessError(
                f"{destination.name} does not have port {destination_port}"
            )

        if (
            destination in self._connections[origin].destinations[origin_port]
            and
            destination_port in self._connections[origin].destinations[origin_port][destination]
            and
            origin in self._connections[destination].origins[destination_port]
            and
            origin_port in self._connections[destination].origins[destination_port][origin]
        ):
            return True

        return False

    def check_connections(self) -> bool:
        """
        Validate that units are connected correctly.

        Warning:
        -------
            If Inlets have ingoing streams.
            If Outlets have outgoing streams.
            If Units (other than Cstr) are not fully connected.

        Returns
        -------
        flag : bool
            True if all units are connected correctly. False otherwise.
        """
        flag = True
        for unit, connections in self.connections.items():
            if isinstance(unit, Inlet):
                if connections.origins is not None:
                    flag = False
                    warn("Inlet unit cannot have ingoing stream.")
                if len(connections.destinations) == 0:
                    flag = False
                    warn(f" Unit '{unit.name}' does not have outgoing stream.")
            elif isinstance(unit, Outlet):
                if connections.destinations is not None:
                    flag = False
                    warn("Outlet unit cannot have outgoing stream.")
                if len(connections.origins) == 0:
                    flag = False
                    warn(f"Unit '{unit.name}' does not have ingoing stream.")
            elif isinstance(unit, Cstr):
                if unit.flow_rate is not None and len(connections.destinations) == 0:
                    flag = False
                    warn("Cstr cannot have flow rate without outgoing stream.")
            else:
                if all(
                    len(port_list) == 0 for port_list in connections.origins.values()
                ):
                    flag = False
                    warn(f"Unit '{unit.name}' does not have ingoing stream.")
                if all(
                    len(port_list) == 0
                    for port_list in connections.destinations.values()
                ):
                    flag = False
                    warn(f" Unit '{unit.name}' does not have outgoing stream.")

        return flag

    @property
    def missing_parameters(self) -> list[str]:
        """dict: Missing parameters of the flow sheet."""
        missing_parameters = []

        for unit in self.units:
            missing_parameters += [
                f"{unit.name}.{param}" for param in unit.missing_parameters
            ]

        return missing_parameters

    def check_units_config(self) -> bool:
        """
        Check if units are configured correctly.

        Returns
        -------
        flag : bool
            True if units are configured correctly. False otherwise.
        """
        flag = True
        for unit in self.units:
            if not unit.check_required_parameters():
                flag = False
        return flag

    @property
    def output_states(self) -> Dict:
        """dict: Output states of the unit operations."""
        output_states_dict = self._output_states.copy()

        for unit, ports in output_states_dict.items():
            for port in ports:
                if port is None:
                    output_states_dict[unit] = output_states_dict[unit][port]

        return output_states_dict

    @unit_name_decorator
    @update_parameters_decorator
    def set_output_state(
        self,
        unit: UnitBaseClass,
        state: int | list[float] | dict,
        port: Optional[str] = None,
    ) -> None:
        """
        Set split ratio of outgoing streams for UnitOperation.

        Parameters
        ----------
        unit : UnitBaseClass
            UnitOperation of flowsheet.
        state : int or list of floats or dict
            new output state of the unit.
        port : str
            Port for which to set the output state.

        Raises
        ------
        CADETProcessError
            If unit not in FlowSheet
            If state is integer and the state >= the state_length.
            If the length of the states is unequal the state_length.
            If the sum of the states is not equal to 1.
            If port cannot be found in the unit operation.
        """

        def get_port_index(
            unit_connection_dict: Dict,
            destination: UnitBaseClass,
            destination_port: str,
        ) -> int:
            """
            Compute the index of a connection for the output state.

            Parameters
            ----------
            unit_connection_dict : Defaultdict
                contains dict with connected units and their respective ports
            destination : UnitBaseClass
                destination object
            destination_port : str
                destination Port
            """
            ret_index = 0
            for unit_destination in unit_connection_dict:
                if unit_destination is destination:
                    ret_index += unit_connection_dict[unit_destination].index(
                        destination_port
                    )
                    break
                ret_index += len(unit_connection_dict[unit_destination])
            return ret_index

        if unit not in self._units:
            raise CADETProcessError("Unit not in flow sheet")

        if port not in unit.ports:
            raise CADETProcessError(f"Port {port} is not a port of Unit {unit.name}")

        state_length = sum([
            len(self._connections[unit].destinations[port][unit_name])
            for unit_name in self._connections[unit].destinations[port]
        ])

        if state_length == 0:
            output_state = []

        if isinstance(state, (int, np.integer)):
            if state >= state_length:
                raise CADETProcessError("Index exceeds destinations")

            output_state = [0.0] * state_length
            output_state[state] = 1.0

        elif isinstance(state, dict):
            output_state = [0.0] * state_length
            for dest, value in state.items():
                try:
                    for destination_port, output_value in value.items():
                        if not self.connection_exists(unit, dest, port, destination_port):
                            raise CADETProcessError(
                                f"{unit} on port {port} does not connect to {dest} on "
                                f"port {destination_port}."
                            )
                        inner_dest = self[dest]
                        index = get_port_index(
                            self._connections[unit].destinations[port],
                            inner_dest,
                            destination_port,
                        )
                        output_state[index] = output_value
                except AttributeError:
                    destination_port = None

                    if not self.connection_exists(unit, dest, port, destination_port):
                        raise CADETProcessError(
                            f"{unit} on port {port} does not connect to {dest} on port "
                            f"{destination_port}."
                        )
                    dest = self[dest]
                    index = get_port_index(
                        self.connections[unit].destinations[port],
                        dest,
                        destination_port,
                    )
                    output_state[index] = value

        elif isinstance(state, (list)):
            if len(state) != state_length:
                raise CADETProcessError(f"Expected length {state_length}.")

            output_state = state
        elif isinstance(state, np.ndarray):
            if len(state) != state_length:
                raise CADETProcessError(f"Expected length {state_length}.")

            output_state = list(state)

        else:
            raise TypeError("Output state must be integer, list or dict.")

        if state_length != 0 and not np.isclose(sum(output_state), 1):
            raise CADETProcessError("Sum of fractions must be 1")

        self._output_states[unit][port] = output_state

    def get_flow_rates(self, state: Optional[Dict] = None, eps: float = 5.9e16) -> Dict:
        r"""
        Calculate flow rate for all connections.

        Optionally, an additional output state can be passed to update the
        current output states.

        Parameters
        ----------
        state : Dict, optional
            Updated flow rates and output states for process sections.
            Default is None.
        eps : float, optional
            eps as an upper boarder for condition of flow_rate calculation

        Returns
        -------
        Dict
            Volumetric flow rate for each unit operation.

        Notes
        -----
        To calculate the flow rates, a system of equations is set up:

        .. math::

            Q_i = \sum_{j=1}^{n_{units}} q_{ji} = \sum_{j=1}^{n_{units}} Q_j * w_{ji},

        where :math:`Q_i` is the total flow exiting unit :math:`i`, and :math:`w_{ij}`
        is the percentile of the total flow of unit :math:`j` directed to unit
        :math:`i`. If the unit is an `Inlet` or a `Cstr` with a given flow rate,
        :math:`Q_i` is given and the system is simplified. This system is solved using
        `numpy.linalg.solve`. Then, the individual flows :math:`q_{ji}` are extracted.

        Raises
        ------
        CADETProcessError
            If flow sheet connectivity matrix is singular, indicating a potential issue
            in flow sheet configuration.

        References
        ----------
        Forum discussion on flow rate calculation:
        https://forum.cadet-web.de/t/improving-the-flowrate-calculation/795
        """
        port_index_list = []
        port_number = 0

        for unit in self.units:
            for port in unit.ports:
                port_index_list.append((unit, port))
                port_number += 1

        flow_rates = {
            unit.name: unit.flow_rate
            for unit in (self.inlets + self.cstrs)
            if unit.flow_rate is not None
        }

        output_states = self._output_states

        if state is not None:
            for param, value in state.items():
                param = param.split(".")
                param_name = param[-1]
                if param_name == "flow_rate":
                    unit_name = param[1]
                    flow_rates[unit_name] = value[0]
                elif param[1] == "output_states":
                    unit_name = param[2]
                    unit = self.units_dict[unit_name]
                    if unit.has_ports:
                        port = param[2]
                    else:
                        port = None
                    output_states[unit][port] = list(value.ravel())

        # Setup matrix with output states.
        w_out = np.zeros((port_number, port_number))

        for unit in self.units:
            if unit.name in flow_rates:
                unit_index = port_index_list.index((unit, None))
                w_out[unit_index, unit_index] = 1
            else:
                for port in self._connections[unit]["origins"]:
                    port_index = port_index_list.index((unit, port))

                    for origin_unit in self._connections[unit]["origins"][port]:
                        for origin_port in self._connections[unit]["origins"][port][origin_unit]:
                            o_index = port_index_list.index((origin_unit, origin_port))
                            local_d_index = 0

                            for inner_unit in self._connections[origin_unit][
                                "destinations"
                            ][origin_port]:
                                if inner_unit == unit:
                                    break
                                local_d_index += len(list(
                                    self._connections[origin_unit]["destinations"][origin_port][inner_unit]
                                ))

                            local_d_index += \
                                self._connections[origin_unit]["destinations"][origin_port][unit].index(port)

                            if output_states[origin_unit][origin_port][local_d_index]:
                                w_out[port_index, o_index] = \
                                    output_states[origin_unit][origin_port][local_d_index]

                    w_out[port_index, port_index] += -1

        # Check for a singular matrix before the loop
        if np.linalg.cond(w_out) > eps:
            raise CADETProcessError(
                "Flow sheet connectivity matrix is singular, which may be due to "
                "unconnected units or missing flow rates. Please ensure all units are "
                "correctly connected and all necessary flow rates are set."
            )

        # Solve system of equations for each polynomial coefficient
        total_flow_rate_coefficents = np.zeros((4, port_number))
        for i in range(4):
            if len(flow_rates) == 0:
                continue

            coeffs = np.array(list(flow_rates.values()), ndmin=2)[:, i]
            if not np.any(coeffs):
                continue

            Q_vec = np.zeros(port_number)
            for unit_name in flow_rates:
                port_index = port_index_list.index((self.units_dict[unit_name], None))
                Q_vec[port_index] = flow_rates[unit_name][i]
            try:
                total_flow_rate_coefficents[i, :] = np.linalg.solve(w_out, Q_vec)
            except np.linalg.LinAlgError:
                raise CADETProcessError(
                    "Unexpected error in flow rate calculation. "
                    "Please check the flow sheet setup."
                )

        # w_out_help is the same as w_out but it contains the origin flow for every unit
        w_out_help = np.zeros((port_number, port_number))

        for unit in self.units:
            if self._connections[unit]["origins"]:
                for port in self._connections[unit]["origins"]:
                    port_index = port_index_list.index((unit, port))

                    for origin_unit in self._connections[unit]["origins"][port]:
                        for origin_port in self._connections[unit]["origins"][port][origin_unit]:
                            o_index = port_index_list.index((origin_unit, origin_port))
                            local_d_index = 0

                            for inner_unit in self._connections[origin_unit]["destinations"][origin_port]:  # noqa: E501
                                if inner_unit == unit:
                                    break
                                local_d_index += len(list(
                                    self._connections[origin_unit]["destinations"][origin_port][inner_unit]
                                ))

                            local_d_index += \
                                self._connections[origin_unit]["destinations"][origin_port][unit].index(port)

                            if not output_states[origin_unit][origin_port][local_d_index]:
                                w_out_help[port_index, o_index] = 0
                            else:
                                w_out_help[port_index, o_index] = \
                                    output_states[origin_unit][origin_port][local_d_index]

        # Calculate total_in as a matrix in "one" step rather than iterating manually.
        total_in_matrix = w_out_help @ total_flow_rate_coefficents.T

        # Generate output dict
        return_flow_rates = Dict()
        for unit in self.units:
            unit_solution_dict = Dict()

            if not isinstance(unit, Inlet):
                unit_solution_dict["total_in"] = Dict()

                for unit_port in unit.ports:
                    index = port_index_list.index((unit, unit_port))
                    unit_solution_dict["total_in"][unit_port] = \
                        list(total_in_matrix[index])

            if not isinstance(unit, Outlet):
                unit_solution_dict["total_out"] = Dict()

                for unit_port in unit.ports:
                    index = port_index_list.index((unit, unit_port))
                    unit_solution_dict["total_out"][unit_port] = \
                        list(total_flow_rate_coefficents[:, index])

            if not isinstance(unit, Inlet):
                unit_solution_dict["origins"] = Dict()

                for unit_port in self._connections[unit]["origins"]:
                    if self._connections[unit]["origins"][unit_port]:
                        unit_solution_dict["origins"][unit_port] = Dict()

                        for origin_unit in self._connections[unit]["origins"][unit_port]:
                            unit_solution_dict["origins"][unit_port][origin_unit.name] = Dict()

                            for origin_port in self._connections[unit]["origins"][unit_port][origin_unit]:  # noqa: E501
                                origin_port_index = \
                                    port_index_list.index((origin_unit, origin_port))
                                unit_port_index = port_index_list.index((unit, unit_port))
                                flow_list = list(
                                    total_flow_rate_coefficents[:, origin_port_index]
                                    * w_out_help[unit_port_index, origin_port_index]
                                )
                                unit_solution_dict["origins"][unit_port][origin_unit.name][origin_port] = flow_list  # noqa: E501

            if not isinstance(unit, Outlet):
                unit_solution_dict["destinations"] = Dict()

                for unit_port in self._connections[unit]["destinations"]:
                    if self._connections[unit]["destinations"][unit_port]:
                        unit_solution_dict["destinations"][unit_port] = Dict()

                        for destination_unit in self._connections[unit]["destinations"][unit_port]:
                            unit_solution_dict["destinations"][unit_port][destination_unit.name] = Dict()  # noqa 501

                            for destination_port in self._connections[unit]["destinations"][unit_port][destination_unit]:  # noqa E501
                                destination_port_index = port_index_list.index((destination_unit, destination_port))  # noqa 501
                                unit_port_index = port_index_list.index((unit, unit_port))
                                flow_list = list(
                                    total_flow_rate_coefficents[:, unit_port_index]
                                    * w_out_help[destination_port_index, unit_port_index]
                                )
                                unit_solution_dict["destinations"][unit_port][destination_unit.name][destination_port] = flow_list  # noqa 501

            return_flow_rates[unit.name] = unit_solution_dict

        return return_flow_rates

    def check_flow_rates(self, state: Optional[dict] = None) -> None:
        """Check if in and outgoing flow rates of unit operations are balanced."""
        flow_rates = self.get_flow_rates(state)
        for unit, q in flow_rates.items():
            if isinstance(unit, (Inlet, Outlet)):
                continue
            elif isinstance(unit, Cstr) and Cstr.flow_rate is not None:
                continue

            if not np.all(q.total_in == q.total_out):
                raise CADETProcessError(f"Unbalanced flow rate for unit '{unit}'.")

    @property
    def feed_inlets(self) -> list[UnitBaseClass]:
        """list: Inlets considered for calculating recovery yield."""
        return self._feed_inlets

    @unit_name_decorator
    def add_feed_inlet(self, feed_inlet: SourceMixin) -> None:
        """
        Add inlet to list of units to be considered for recovery.

        Parameters
        ----------
        feed_inlet : SourceMixin
            Unit to be added to list of feed inlets.

        Raises
        ------
        CADETProcessError
            If unit is not an Inlet.
            If unit is already marked as feed inlet.
        """
        if feed_inlet not in self.inlets:
            raise CADETProcessError("Expected Inlet")
        if feed_inlet in self._feed_inlets:
            raise CADETProcessError(f"Unit '{feed_inlet}' is already a feed inlet.")
        self._feed_inlets.append(feed_inlet)

    @unit_name_decorator
    def remove_feed_inlet(self, feed_inlet: SourceMixin) -> None:
        """
        Remove inlet from list of units to be considered for recovery.

        Parameters
        ----------
        feed_inlet : SourceMixin
            Unit to be removed from list of feed inlets.
        """
        if feed_inlet not in self._feed_inlets:
            raise CADETProcessError(f"Unit '{feed_inlet}' is not a feed inlet.")
        self._feed_inlets.remove(feed_inlet)

    @property
    def eluent_inlets(self) -> list[UnitBaseClass]:
        """list: Inlets to be considered for eluent consumption."""
        return self._eluent_inlets

    @unit_name_decorator
    def add_eluent_inlet(self, eluent_inlet: SourceMixin) -> None:
        """
        Add inlet to list of units to be considered for eluent consumption.

        Parameters
        ----------
        eluent_inlet : SourceMixin
            Unit to be added to list of eluent inlets.

        Raises
        ------
        CADETProcessError
            If unit is not an Inlet.
            If unit is already marked as eluent inlet.
        """
        if eluent_inlet not in self.inlets:
            raise CADETProcessError("Expected Inlet")
        if eluent_inlet in self._eluent_inlets:
            raise CADETProcessError(f"Unit '{eluent_inlet}' is already an eluent inlet")
        self._eluent_inlets.append(eluent_inlet)

    @unit_name_decorator
    def remove_eluent_inlet(self, eluent_inlet: SourceMixin) -> None:
        """
        Remove inlet from list of units considered for eluent consumption.

        Parameters
        ----------
        eluent_inlet : SourceMixin
            Unit to be added to list of eluent inlets.

        Raises
        ------
        CADETProcessError
            If unit is not in eluent inlets.
        """
        if eluent_inlet not in self._eluent_inlets:
            raise CADETProcessError(f"Unit '{eluent_inlet}' is not an eluent inlet.")
        self._eluent_inlets.remove(eluent_inlet)

    @property
    def product_outlets(self) -> list[UnitBaseClass]:
        """list: Outlets to be considered for fractionation."""
        return self._product_outlets

    @unit_name_decorator
    def add_product_outlet(self, product_outlet: Outlet) -> None:
        """
        Add outlet to list of units considered for fractionation.

        Parameters
        ----------
        product_outlet : Outlet
            Unit to be added to list of product outlets.

        Raises
        ------
        CADETProcessError
            If unit is not an Outlet.
            If unit is already marked as product outlet.
        """
        if product_outlet not in self.outlets:
            raise CADETProcessError("Expected Outlet")
        if product_outlet in self._product_outlets:
            raise CADETProcessError(
                f"Unit '{product_outlet}' is already a product outlet"
            )
        self._product_outlets.append(product_outlet)

    @unit_name_decorator
    def remove_product_outlet(self, product_outlet: Outlet) -> None:
        """
        Remove outlet from list of units to be considered for fractionation.

        Parameters
        ----------
        product_outlet : Outlet
            Unit to be added to list of product outlets.

        Raises
        ------
        CADETProcessError
            If unit is not a product outlet.
        """
        if product_outlet not in self._product_outlets:
            raise CADETProcessError(f"Unit '{product_outlet}' is not a product outlet.")
        self._product_outlets.remove(product_outlet)

    @property
    def parameters(self) -> dict:
        """dict: Parameters of the flow sheet and associated unit operations."""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        try:
            output_states = parameters.pop("output_states")
            for unit, state in output_states.items():
                unit = self.units_dict[unit]
                # Hier, if unit.has_ports: iterate.
                if not unit.has_ports:
                    self.set_output_state(unit, state)
                else:
                    for port_i, state_i in state.items():
                        self.set_output_state(unit, state_i, port_i)
        except KeyError:
            pass

        for unit, params in parameters.items():
            if unit not in self.units_dict:
                raise CADETProcessError("Not a valid unit")
            self.units_dict[unit].parameters = params

        self.update_parameters()

    @property
    def sized_parameters(self) -> list[str]:
        """list: List of sized parameters."""
        return self._sized_parameters

    @property
    def polynomial_parameters(self) -> list[str]:
        """list: List of polynomial parameters."""
        return self._polynomial_parameters

    @property
    def section_dependent_parameters(self) -> list[str]:
        """list: List of section dependent parameters."""
        return self._section_dependent_parameters

    @property
    def initial_state(self) -> dict:
        """dict: Initial state of the unit oeprations."""
        initial_state = {unit.name: unit.initial_state for unit in self.units}

        return initial_state

    @initial_state.setter
    def initial_state(self, initial_state: dict) -> None:
        for unit, st in initial_state.items():
            if unit not in self.units_dict:
                raise CADETProcessError("Not a valid unit")
            self.units_dict[unit].initial_state = st

    def __len__(self) -> int:
        """int: The number of unit operations in the system."""
        return self.number_of_units

    def __iter__(self) -> Iterator[UnitBaseClass]:
        """Iterate over the unit operations in the system."""
        return iter(self.units)

    def __getitem__(self, unit_name: str) -> UnitBaseClass:
        """
        Make FlowSheet substriptable s.t. units can be used as keys.

        Parameters
        ----------
        unit_name : str
            Name of the unit.

        Returns
        -------
        unit : UnitBaseClass
            UnitOperation of FlowSheet.

        Raises
        ------
        KeyError
            If unit not in FlowSheet
        """
        try:
            return self.units_dict[unit_name]
        except KeyError:
            raise KeyError("Not a valid unit")

    def __contains__(self, item: UnitBaseClass | str) -> bool:
        """
        Check if UnitOperation is part of the FlowSheet.

        Parameters
        ----------
        item : UnitBaseClass
            item to be searched

        Returns
        -------
        Bool : True if item is in units, otherwise False.
        """
        if (item in self._units) or (item in self.unit_names):
            return True
        else:
            return False



import numpy as np
import pytest
from CADETProcess.processModel import (
    MCT,
    ComponentSystem,
    Cstr,
    GeneralRateModel,
    Inlet,
    LumpedRateModelWithoutPores,
    LumpedRateModelWithPores,
    TubularReactor,
)

length = 0.6
diameter = 0.024

cross_section_area = np.pi / 4 * diameter**2
volume_liquid = cross_section_area * length
volume = cross_section_area * length

bed_porosity = 0.3
particle_porosity = 0.6
particle_radius = 1e-4
total_porosity = bed_porosity + (1 - bed_porosity) * particle_porosity
const_solid_volume = volume * (1 - total_porosity)
init_liquid_volume = volume * total_porosity

axial_dispersion = 4.7e-7
film_diffusion_0 = 0
film_diffusion_1 = 1e-6
pore_diffusion_0 = 0
pore_diffusion_1 = 1e-11

nchannel = 3
channel_cross_section_areas = [0.1, 0.1, 0.1]
exchange_matrix = np.array([
    [[0.0], [0.01], [0.0]],
    [[0.02], [0.0], [0.03]],
    [[0.0], [0.0], [0.0]]
])
flow_direction = 1


@pytest.fixture
def component_system():
    return ComponentSystem(2)


@pytest.fixture
def inlet(component_system):
    return Inlet(component_system, name="test_inlet")


@pytest.fixture
def cstr(component_system):
    cstr = Cstr(component_system, name="test_cstr")
    cstr.const_solid_volume = const_solid_volume
    cstr.init_liquid_volume = init_liquid_volume
    cstr.flow_rate = 1
    return cstr


@pytest.fixture
def tubular_reactor(component_system):
    tubular_reactor = TubularReactor(component_system, name="test_tubular_reactor")
    tubular_reactor.length = length
    tubular_reactor.diameter = diameter
    tubular_reactor.axial_dispersion = axial_dispersion
    return tubular_reactor


@pytest.fixture
def lrm(component_system):
    lrm = LumpedRateModelWithoutPores(component_system, name="test_lrm")
    lrm.length = length
    lrm.diameter = diameter
    lrm.axial_dispersion = axial_dispersion
    lrm.total_porosity = total_porosity
    return lrm


@pytest.fixture
def lrmp(component_system):
    lrmp = LumpedRateModelWithPores(component_system, name="test_lrmp")
    lrmp.length = length
    lrmp.diameter = diameter
    lrmp.axial_dispersion = axial_dispersion
    lrmp.bed_porosity = bed_porosity
    lrmp.particle_radius = particle_radius
    lrmp.particle_porosity = particle_porosity
    lrmp.film_diffusion = [film_diffusion_0, film_diffusion_1]
    return lrmp


@pytest.fixture
def grm(components=2):
    grm = GeneralRateModel(ComponentSystem(components), name="test_grm")
    grm.length = length
    grm.diameter = diameter
    grm.axial_dispersion = axial_dispersion
    grm.bed_porosity = bed_porosity
    grm.particle_radius = particle_radius
    grm.particle_porosity = particle_porosity
    grm.film_diffusion = [film_diffusion_0, film_diffusion_1]
    grm.pore_diffusion = [pore_diffusion_0, pore_diffusion_1]
    return grm


@pytest.fixture
def mct(components=1):
    mct = MCT(ComponentSystem(components), nchannel=3, name="test_mct")
    mct.length = length
    mct.channel_cross_section_areas = channel_cross_section_areas
    mct.axial_dispersion = axial_dispersion
    mct.exchange_matrix = exchange_matrix
    return mct


@pytest.mark.parametrize(
    "unit_operation, expected_geometry",
    [
        (
            "cstr",
            {
                "volume_liquid": total_porosity * volume,
                "volume_solid": (1 - total_porosity) * volume,
            },
        ),
        (
            "lrm",
            {
                "cross_section_area": cross_section_area,
                "total_porosity": total_porosity,
                "volume": volume,
                "volume_interstitial": total_porosity * volume,
                "volume_liquid": total_porosity * volume,
                "volume_solid": (1 - total_porosity) * volume,
            },
        ),
        (
            "lrmp",
            {
                "cross_section_area": cross_section_area,
                "total_porosity": total_porosity,
                "volume": volume,
                "volume_interstitial": bed_porosity * volume,
                "volume_liquid": total_porosity * volume,
                "volume_solid": (1 - total_porosity) * volume,
            },
        ),
        (
            "mct",
            {
                "volume": length * sum(channel_cross_section_areas),
            },
        ),
    ],
)
def test_geometry(unit_operation, expected_geometry, request):
    unit = request.getfixturevalue(unit_operation)

    if "total_porosity" in expected_geometry:
        assert unit.total_porosity == expected_geometry["total_porosity"]

    if "volume" in expected_geometry:
        assert unit.volume == expected_geometry["volume"]

    if "volume_interstitial" in expected_geometry:
        assert np.isclose(
            unit.volume_interstitial, expected_geometry["volume_interstitial"]
        )

    if "volume_liquid" in expected_geometry:
        assert np.isclose(unit.volume_liquid, expected_geometry["volume_liquid"])

    if "volume_solid" in expected_geometry:
        assert np.isclose(unit.volume_solid, expected_geometry["volume_solid"])

    if "cross_section_area" in expected_geometry:
        assert unit.cross_section_area == expected_geometry["cross_section_area"]

        unit.cross_section_area = cross_section_area / 2
        assert np.isclose(unit.diameter, diameter / (2**0.5))


@pytest.mark.parametrize(
    "input_c, expected_c",
    [
        (1, np.array([[1, 0, 0, 0], [1, 0, 0, 0]])),
        ([1, 1], np.array([[1, 0, 0, 0], [1, 0, 0, 0]])),
        ([1, 2], np.array([[1, 0, 0, 0], [2, 0, 0, 0]])),
        ([[1, 0], [2, 0]], np.array([[1, 0, 0, 0], [2, 0, 0, 0]])),
        ([[1, 2], [3, 4]], np.array([[1, 2, 0, 0], [3, 4, 0, 0]])),
    ],
)
def test_polynomial_inlet_concentration(inlet, input_c, expected_c):
    inlet.c = input_c
    np.testing.assert_equal(inlet.c, expected_c)


@pytest.mark.parametrize(
    "unit_operation, input_flow_rate, expected_flow_rate",
    [
        ("inlet", 1, np.array([1, 0, 0, 0])),
        ("inlet", [1, 0], np.array([1, 0, 0, 0])),
        ("inlet", [1, 1], np.array([1, 1, 0, 0])),
        ("cstr", 1, np.array([1, 0, 0, 0])),
        ("cstr", [1, 0], np.array([1, 0, 0, 0])),
        ("cstr", [1, 1], np.array([1, 1, 0, 0])),
    ],
)
def test_polynomial_flow_rate(
    unit_operation, input_flow_rate, expected_flow_rate, request
):
    unit = request.getfixturevalue(unit_operation)
    unit.flow_rate = input_flow_rate
    np.testing.assert_equal(unit.flow_rate, expected_flow_rate)


@pytest.mark.parametrize(
    "unit_operation, expected_parameters",
    [
        (
            "cstr",
            {
                "flow_rate": np.array([1, 0, 0, 0]),
                "init_liquid_volume": init_liquid_volume,
                "flow_rate_filter": 0,
                "c": [0, 0],
                "q": None,
                "const_solid_volume": const_solid_volume,
            },
        ),
        (
            "tubular_reactor",
            {
                "length": length,
                "diameter": diameter,
                "axial_dispersion": [axial_dispersion, axial_dispersion],
                "flow_direction": flow_direction,
                "c": [0, 0],
                "discretization": {
                    "ncol": 100,
                    "use_analytic_jacobian": True,
                    "reconstruction": "WENO",
                    "weno": {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
                    "consistency_solver": {
                        "solver_name": "LEVMAR",
                        "init_damping": 0.01,
                        "min_damping": 0.0001,
                        "max_iterations": 50,
                        "subsolvers": "LEVMAR",
                    },
                },
            },
        ),
        (
            "lrm",
            {
                "length": length,
                "diameter": diameter,
                "total_porosity": total_porosity,
                "axial_dispersion": [axial_dispersion, axial_dispersion],
                "flow_direction": flow_direction,
                "c": [0, 0],
                "q": None,
                "discretization": {
                    "ncol": 100,
                    "use_analytic_jacobian": True,
                    "reconstruction": "WENO",
                    "weno": {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
                    "consistency_solver": {
                        "solver_name": "LEVMAR",
                        "init_damping": 0.01,
                        "min_damping": 0.0001,
                        "max_iterations": 50,
                        "subsolvers": "LEVMAR",
                    },
                },
            },
        ),
        (
            "lrmp",
            {
                "length": length,
                "diameter": diameter,
                "bed_porosity": bed_porosity,
                "axial_dispersion": [axial_dispersion, axial_dispersion],
                "pore_accessibility": [1, 1],
                "film_diffusion": [film_diffusion_0, film_diffusion_1],
                "particle_radius": particle_radius,
                "particle_porosity": particle_porosity,
                "flow_direction": flow_direction,
                "c": [0, 0],
                "cp": [0, 0],
                "q": None,
                "discretization": {
                    "ncol": 100,
                    "par_geom": "SPHERE",
                    "use_analytic_jacobian": True,
                    "gs_type": True,
                    "max_krylov": 0,
                    "max_restarts": 10,
                    "schur_safety": 1e-08,
                    "reconstruction": "WENO",
                    "weno": {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
                    "consistency_solver": {
                        "solver_name": "LEVMAR",
                        "init_damping": 0.01,
                        "min_damping": 0.0001,
                        "max_iterations": 50,
                        "subsolvers": "LEVMAR",
                    },
                },
            },
        ),
        (
            "grm",
            {
                "length": length,
                "diameter": diameter,
                "bed_porosity": bed_porosity,
                "axial_dispersion": [axial_dispersion, axial_dispersion],
                "pore_accessibility": [1, 1],
                "film_diffusion": [film_diffusion_0, film_diffusion_1],
                "particle_radius": particle_radius,
                "particle_porosity": particle_porosity,
                "pore_diffusion": [pore_diffusion_0, pore_diffusion_1],
                "surface_diffusion": None,
                "flow_direction": flow_direction,
                "c": [0, 0],
                "cp": [0, 0],
                "q": None,
                "discretization": {
                    "ncol": 100,
                    "par_geom": "SPHERE",
                    "npar": 5,
                    "par_disc_type": "EQUIDISTANT_PAR",
                    "par_boundary_order": 2,
                    "fix_zero_surface_diffusion": False,
                    "use_analytic_jacobian": True,
                    "gs_type": True,
                    "max_krylov": 0,
                    "max_restarts": 10,
                    "schur_safety": 1e-08,
                    "reconstruction": "WENO",
                    "weno": {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
                    "consistency_solver": {
                        "solver_name": "LEVMAR",
                        "init_damping": 0.01,
                        "min_damping": 0.0001,
                        "max_iterations": 50,
                        "subsolvers": "LEVMAR",
                    },
                },
            },
        ),
        (
            "mct",
            {
                "nchannel": nchannel,
                "length": length,
                "channel_cross_section_areas": channel_cross_section_areas,
                "axial_dispersion": nchannel * [axial_dispersion],
                "exchange_matrix": exchange_matrix,
                "flow_direction": 1,
                "c": [[0, 0, 0]],
                "discretization": {
                    "ncol": 100,
                    "use_analytic_jacobian": True,
                    "reconstruction": "WENO",
                    "weno": {"boundary_model": 0, "weno_eps": 1e-10, "weno_order": 3},
                    "consistency_solver": {
                        "solver_name": "LEVMAR",
                        "init_damping": 0.01,
                        "min_damping": 0.0001,
                        "max_iterations": 50,
                        "subsolvers": "LEVMAR",
                    },
                },
            },
        ),
    ],
)
def test_parameters(unit_operation, expected_parameters, request):
    unit = request.getfixturevalue(unit_operation)
    np.testing.assert_equal(expected_parameters, unit.parameters)


@pytest.mark.parametrize(
    "unit_operation, flow_rate, expected_velocity",
    [
        ("tubular_reactor", 2, 2),
        ("lrmp", 2, 4),
        ("tubular_reactor", 0, ZeroDivisionError),
        ("lrmp", 0, ZeroDivisionError),
    ],
)
def test_interstitial_velocity(unit_operation, flow_rate, expected_velocity, request):
    unit = request.getfixturevalue(unit_operation)
    unit.length = 1
    unit.cross_section_area = 1
    unit.axial_dispersion = 3 if unit_operation == "tubular_reactor" else [3, 2]

    if unit_operation == "lrmp":
        unit.bed_porosity = 0.5

    if expected_velocity is ZeroDivisionError:
        with pytest.raises(ZeroDivisionError):
            unit.calculate_interstitial_velocity(flow_rate)
    else:
        assert np.isclose(
            unit.calculate_interstitial_velocity(flow_rate), expected_velocity
        )


@pytest.mark.parametrize(
    "unit_operation, flow_rate, expected_velocity",
    [
        ("tubular_reactor", 2, 2),
        ("lrmp", 2, 2),
        ("tubular_reactor", 0, ZeroDivisionError),
        ("lrmp", 0, ZeroDivisionError),
    ],
)
def test_superficial_velocity(unit_operation, flow_rate, expected_velocity, request):
    unit = request.getfixturevalue(unit_operation)
    unit.length = 1
    unit.cross_section_area = 1

    if unit_operation == "lrmp":
        unit.bed_porosity = 0.5

    if expected_velocity is ZeroDivisionError:
        with pytest.raises(ZeroDivisionError):
            unit.calculate_superficial_velocity(flow_rate)
    else:
        assert np.isclose(
            unit.calculate_superficial_velocity(flow_rate), expected_velocity
        )


@pytest.mark.parametrize(
    "unit_operation, flow_rate, expected_ntp",
    [
        ("tubular_reactor", 2, [1 / 3, 1 / 3]),
        ("tubular_reactor", 0, ZeroDivisionError),
    ],
)
def test_ntp(unit_operation, flow_rate, expected_ntp, request):
    unit = request.getfixturevalue(unit_operation)
    unit.length = 1
    unit.cross_section_area = 1
    unit.axial_dispersion = 3

    if expected_ntp is ZeroDivisionError:
        with pytest.raises(ZeroDivisionError):
            unit.NTP(flow_rate)
    else:
        np.testing.assert_almost_equal(unit.NTP(flow_rate), expected_ntp)


@pytest.mark.parametrize(
    "unit_operation, flow_rate, axial_dispersion, expected_bodenstein",
    [
        ("tubular_reactor", 2, 3, [2 / 3, 2 / 3]),
        ("lrmp", 2, [3, 2], [4 / 3, 2]),
        ("lrmp", 2, [1, 2], [4, 2]),
    ],
)
def test_bodenstein_number(
    unit_operation, flow_rate, axial_dispersion, expected_bodenstein, request
):
    unit = request.getfixturevalue(unit_operation)
    unit.length = 1
    unit.cross_section_area = 1
    unit.axial_dispersion = axial_dispersion

    if unit_operation == "lrmp":
        unit.bed_porosity = 0.5

    np.testing.assert_almost_equal(
        unit.calculate_bodenstein_number(flow_rate), expected_bodenstein
    )


if __name__ == "__main__":
    pytest.main([__file__])



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# (lwe_example_flow_rate)=
# # Flow Rate Gradients
#
#
# ```{figure} ./figures/flow_sheet_flow_rate.svg
# Flow sheet for load-wash-elute process using a separate inlets for buffers.
# ```

# %%
from CADETProcess.processModel import ComponentSystem
from CADETProcess.processModel import StericMassAction
from CADETProcess.processModel import Inlet, GeneralRateModel, Outlet
from CADETProcess.processModel import FlowSheet
from CADETProcess.processModel import Process

# Component System
component_system = ComponentSystem()
component_system.add_component('Salt')
component_system.add_component('A')
component_system.add_component('B')
component_system.add_component('C')

# Binding Model
binding_model = StericMassAction(component_system, name='SMA')
binding_model.is_kinetic = True
binding_model.adsorption_rate = [0.0, 35.5, 1.59, 7.7]
binding_model.desorption_rate = [0.0, 1000, 1000, 1000]
binding_model.characteristic_charge = [0.0, 4.7, 5.29, 3.7]
binding_model.steric_factor = [0.0, 11.83, 10.6, 10]
binding_model.capacity = 1200.0

# Unit Operations
load = Inlet(component_system, name='load')
load.c = [50, 1.0, 1.0, 1.0]

wash = Inlet(component_system, name='wash')
wash.c = [50.0, 0.0, 0.0, 0.0]

elute = Inlet(component_system, name='elute')
elute.c = [500.0, 0.0, 0.0, 0.0]

column = GeneralRateModel(component_system, name='column')
column.binding_model = binding_model

column.length = 0.014
column.diameter = 0.02
column.bed_porosity = 0.37
column.particle_radius = 4.5e-5
column.particle_porosity = 0.75
column.axial_dispersion = 5.75e-8
column.film_diffusion = column.n_comp * [6.9e-6]
column.pore_diffusion = [7e-10, 6.07e-11, 6.07e-11, 6.07e-11]
column.surface_diffusion = column.n_bound_states * [0.0]

column.c = [50, 0, 0, 0]
column.cp = [50, 0, 0, 0]
column.q = [binding_model.capacity, 0, 0, 0]

outlet = Outlet(component_system, name='outlet')

# Flow Sheet
flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(load, feed_inlet=True)
flow_sheet.add_unit(wash, eluent_inlet=True)
flow_sheet.add_unit(elute, eluent_inlet=True)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet, product_outlet=True)

flow_sheet.add_connection(load, column)
flow_sheet.add_connection(wash, column)
flow_sheet.add_connection(elute, column)
flow_sheet.add_connection(column, outlet)

# %% [markdown]
# ```{figure} ./figures/events_flow_rate.svg
# Events of load-wash-elute process using multiple inlets and mofifying their flow rates.
# ```

# %%
# Process
process = Process(flow_sheet, 'lwe')
process.cycle_time = 2000.0

load_duration = 10.0
t_gradient_start = 90.0
gradient_duration = process.cycle_time - t_gradient_start

Q = 6.683738370512285e-8
gradient_slope = Q / (process.cycle_time - t_gradient_start)

process.add_event('load_on', 'flow_sheet.load.flow_rate', Q)
process.add_event('load_off', 'flow_sheet.load.flow_rate', 0.0)
process.add_duration('load_duration', time=load_duration)
process.add_event_dependency('load_off', ['load_on', 'load_duration'], [1, 1])

process.add_event('wash_off', 'flow_sheet.wash.flow_rate', 0)
process.add_event(
    'elute_off', 'flow_sheet.elute.flow_rate', 0
)

process.add_event(
    'wash_on', 'flow_sheet.wash.flow_rate', Q, time=load_duration
)
process.add_event_dependency('wash_on', ['load_off'])

process.add_event(
    'wash_gradient', 'flow_sheet.wash.flow_rate',
    [Q, -gradient_slope], t_gradient_start
    )
process.add_event(
    'elute_gradient', 'flow_sheet.elute.flow_rate', [0, gradient_slope]
    )
process.add_event_dependency('elute_gradient', ['wash_gradient'])

# %%
if __name__ == '__main__':
    from CADETProcess.simulator import Cadet
    process_simulator = Cadet()

    simulation_results = process_simulator.simulate(process)

    from CADETProcess.plotting import SecondaryAxis
    sec = SecondaryAxis()
    sec.components = ['Salt']
    sec.y_label = '$c_{salt}$'

    simulation_results.solution.column.outlet.plot(secondary_axis=sec)
